
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVASpell: component for                        }
{       spell-checking RichViewEdit using               }
{       GNU-Aspell-0.50                                 }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       ASpell : Copyright (c) Kevin Atkinson           }
{       kevina@gnu.org                                  }
{       http://aspell.net                               }
{                                                       }
{*******************************************************}

{

  This component is a wrapper for ASpell DLL.
  This DLL may be distributed under the terms of the GNU Lesser General Public
  License (LGPL): http://www.gnu.org/copyleft/lesser.html
  (Important conditions:
  LGPL allows using ASpell DLL both in free and non-free programs.
  You must give prominent notice with each copy of your application that
  ASpell is used in it and that ASpell and its use are covered by LGPL.
  You must supply a copy of LGPL.
  If your application displays copyright notices, you must include the
  copyright notice for ASpell among them, as well as a reference directing
  the user to the copy of LGPL.
  Besides, you must give your users a possibility to get a source code of
  ASpell DLL (see section 6 of LGPL))


  Installing this component:

    - Add this file (RVASpell.pas) in some package. Compile this package.
      TRVASpell component will appear on "RichView" page of the
      Component Palette.

  Using:
    - place TRVASpell on the form.
    - Check IsDLLLoaded property. If it's False, ASpell DLL cannot be loaded.
      It's reloaded when you assign new value to ASpellDLL property.
    - in program, call TRVASpell.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call TRVASpell.Check(rve, rvesFromStart) to check from the
      beginning.
  Properties:
    - ASpellDLL: TFileName - path and file name of ASpellDLL.
      If it is blank (default), it is read from the Registry
      (HKLM\SOFTWARE\Aspell)
    - Dictionaries: TStringList - list of available dictionaries.
    - DictIndex: Integer - index of active dictionary.
    - IsDLLLoaded: Boolean, read-only
    - SpellFormStyle defines which dialog style is used
      (classic or like in MS Word)
  Events:
    - OnSpellForm - event occuring when spell form is created. Can be used for
      localization.
  This component does not provide advanced configuration of ASpell.
  It can be done using ASpell API (see AspellHeadersDyn.pas and ASpell
  documentation)
  Functions for live spelling:
    - checking word in RichViewEdit.OnSpellingCheck: TRVASpell.IsMisspelled
    - adding to ignore list: TRVASpell.AddToIgnoreList
    - adding to dictionary: TRVASpell.AddToDictionary
    - suggestions: TRVASpell.SuggestWords
    - storing the chosen correction (to generate better suggestions in future):
      TRVASpell.StoreReplacement
    - event on adding word to ignore list: TRVASpell.OnSpellFormAction
    - event on adding word to dictionary: TRVASpell.OnSpellFormAction

  History:
  2011-03-20: property DialogShown specifies if a spelling check dialog has been
    shown during the last spelling check.
  2003-06-06: Ukrainian support
  2004-07-04:
    comments about live spelling
    change: spelling session is started on the component creation and finished on
      the component destroying (previously session was created while calling
      the Check method); the session is restarted when assigning DictIndex
    new: methods and events for live spelling: IsMisspelled, SuggestWords,
      AddToIgnoreList, AddToDictionary, OnSpellFormAction
  2004-08-04:
    chg: Koi8 support is moved to RVKoi8.pas
  2004-08-27:
    fix: no exception if ASpell cannot be loaded any more.
    new: IsDLLLoaded property, StoreReplacement
  2004-09-08:
    fix: no exception if ASpell is installed without dictionaries any more
  2009-02-03:
    new: support for MS Word-like spelling dialog (SpellFormStyle property)
}

unit RVASpell;

interface

{$I RV_Defs.inc}

uses
  SysUtils, Classes, Controls,Forms,
  AspellHeadersDyn,
  RichView, RVEdit, RVKoi8, CRVData, RVUni, RVStyle,
  RVWordEnum, RVCustomSpellFrm, RVTypes;

type
  TRVASpell = class;

  TOnSpellFormEvent = procedure (Sender: TRVASpell; Form: TfrmRVCustomSpell) of object;

  TRVASpellFormAction = (rvasIgnore, rvasIgnoreAll, rvasChange, rvasChangeAll,
    rvasAdd);
  TRVASpellFormActionEvent = procedure(Sender: TRVASpell;
    const AWord, AReplaceTo: String; Action: TRVASpellFormAction) of object;

  TRVASpellFormStyle = (rvafClassic, rvafMSWord);

  TRVASpell = class (TComponent)
    private
      FDLLLoaded: Boolean;
      FASpellDLL: TFileName;
      FDictionaries: TStringList;
      FDictIndex: Integer;
      FOnSpellForm: TOnSpellFormEvent;
      FSpeller: AspellSpeller;
      FOnSpellFormAction: TRVASpellFormActionEvent;
      FRVData: TCustomRVData;
      FSpellFormStyle: TRVASpellFormStyle;
      FDialogShown: Boolean;
      procedure LoadDLL;
      procedure LoadDictList;
      function GetDictName: String;
      function GetDictionaries: TStringList;
      procedure CheckError(possible_err: AspellCanHaveError);
      procedure SetASpellDLL(const Value: TFileName);
      procedure SetDictIndex(const Value: Integer);
      procedure DoSpellFormAction(mr: TModalResult; const AWord, AReplaceTo: String);
      function IsCyrillic: Boolean;
      procedure InitSpell;
      procedure DoneSpell;
    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      property Dictionaries: TStringList read GetDictionaries;
      property DictIndex: Integer read FDictIndex write SetDictIndex;
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      function IsMisspelled(const AWord: String; StyleNo: Integer; rv: TCustomRichView): Boolean;
      procedure SuggestWords(const AWord: String; StyleNo: Integer;
        Suggestions: TStrings; rv: TCustomRichView);
      procedure StoreReplacement(const AWord, ACorrection: String; StyleNo: Integer;
        rv: TCustomRichView);
      procedure AddToDictionary(const AWord: String; StyleNo: Integer; rv: TCustomRichView);
      procedure AddToIgnoreList(const AWord: String; StyleNo: Integer; rv: TCustomRichView);
      property IsDLLLoaded: Boolean read FDLLLoaded;
      property DialogShown: Boolean read FDialogShown;
    published
      property SpellFormStyle: TRVASpellFormStyle
        read FSpellFormStyle write FSpellFormStyle default rvafClassic;
      property ASpellDLL: TFileName read FASpellDLL write SetASpellDLL;
      property OnSpellForm: TOnSpellFormEvent read FOnSpellForm write FOnSpellForm;
      property OnSpellFormAction: TRVASpellFormActionEvent read FOnSpellFormAction
        write FOnSpellFormAction;
  end;

  TRVASpellHelper = class (TRVWordEnumerator)
    private
      FRVASpell: TRVASpell;
      FReplaceFrom, FReplaceTo: TStringList;
      FCyrillic: Boolean;
      frm : TfrmRVCustomSpell;
    protected
      function ProcessWord: Boolean; override;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
    published
      property RVASpell: TRVASpell read FRVASpell write FRVASpell;
  end;

  ERVASpellError = class (Exception);

  procedure Register;

implementation
uses RVSpellFormUtils;


function GetDictCodePage(const DictName: TRVAnsiString): TRVCodePage;
begin
  if (DictName='cs') or (DictName='pl') or (DictName='ro') or (DictName='sl') then
    Result := 1250   // Eastern European
  else if (DictName='el') then
    Result := 1253  // Greek
  else if (DictName='ru') or (DictName='uk') then
    Result := 1251   // Cyrillic
  else
    Result := 1252;
end;
{=================================== TRVASpell ================================}
constructor TRVASpell.Create(AOwner: TComponent);
begin
  inherited;
  FDictionaries := TStringList.Create;
  InitSpell;
end;
{------------------------------------------------------------------------------}
destructor TRVASpell.Destroy;
begin
  DoneSpell;
  FDictionaries.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.SetASpellDLL(const Value: TFileName);
begin
  if FASpellDLL<>Value then begin
    FASpellDLL := Value;
    FDLLLoaded := False;
    LoadDLL;
  end;
end;
{------------------------------------------------------------------------------}
function TRVASpell.GetDictionaries: TStringList;
begin
  LoadDLL;
  Result := FDictionaries;
end;
{------------------------------------------------------------------------------}
function TRVASpell.GetDictName: String;
var SpacePos: Integer;
begin
  Result := '';
  if FDictionaries.Count=0 then
    exit;
  if (FDictIndex<0) or (FDictIndex>=FDictionaries.Count) then
    FDictIndex := 0;
  Result := FDictionaries[FDictIndex];
  SpacePos := Pos(' ', Result);
  if SpacePos>0 then
    Result := Copy(Result, 1, SpacePos-1);
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.LoadDictList;
var
  cnf : AspellConfig;
  dict_info_list : AspellDictInfoList;
  dict_info_elements : AspellDictInfoEnumeration;
  dict_info : TAspellDictInfo;
begin
  FDictionaries.Clear;
  if not Assigned(new_aspell_config) then
    exit;
  cnf := new_aspell_config();
  dict_info_list := get_aspell_dict_info_list(cnf);
  delete_aspell_config(cnf);
  dict_info_elements := aspell_dict_info_list_elements(dict_info_list);
  while aspell_dict_info_enumeration_at_end(dict_info_elements)=0 do begin
    dict_info := aspell_dict_info_enumeration_next( dict_info_elements )^;
    FDictionaries.Add(String(dict_info.name));
  end;
  delete_aspell_dict_info_enumeration(dict_info_elements);
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.LoadDLL;
begin
  if FDLLLoaded then
    exit;
  LoadAspell(ASpellDLL);
  FDLLLoaded := Assigned(new_aspell_config);
  if FDLLLoaded then
    LoadDictList;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.CheckError(possible_err: AspellCanHaveError);
begin
  if (aspell_error_number(possible_err) <> 0) then
    raise Exception.Create(String(aspell_error_message(possible_err)));
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.InitSpell;
var
  cnf : AspellConfig;
  possible_err  : AspellCanHaveError;
  DictName: TRVAnsiString;
begin
  LoadDLL;
  if not FDLLLoaded then
    exit;
  if FSpeller=nil then begin
    cnf := new_aspell_config();
    try
      try
        DictName := TRVAnsiString(GetDictName);
        aspell_config_replace(cnf, 'lang', PRVAnsiChar(DictName));
        possible_err := new_aspell_speller(cnf);
        CheckError(possible_err);
        FSpeller := to_aspell_speller(possible_err);
      finally
        delete_aspell_config(cnf);
      end;
    except
      ;
    end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.DoneSpell;
begin
  if FSpeller=nil then
    exit;
  aspell_speller_save_all_word_lists(FSpeller);
  delete_aspell_speller(FSpeller);
  FSpeller := nil;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.SetDictIndex(const Value: Integer);
begin
  DoneSpell;
  FDictIndex := Value;
  InitSpell;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var Helper: TRVASpellHelper;
begin
  if FSpeller=nil then
    exit;
  Helper := TRVASpellHelper.Create;
  try
    Helper.RVASpell := Self;
    FRVData := Editor.RVData;
    try
      Helper.Check(Editor, Scope);
      FDialogShown := Helper.DialogShown;
    finally
      Helper.FRVData := nil;
    end;
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
function TRVASpell.IsCyrillic: Boolean;
var s: String;
begin
  s := GetDictName;
  Result := (s='ru') or (s='uk');
end;
{------------------------------------------------------------------------------}
function TRVASpell.IsMisspelled(const AWord: String; StyleNo: Integer;
  rv: TCustomRichView): Boolean;
var LWord: TRVRawByteString;
begin
  if FSpeller=nil then begin
    Result := False;
    exit;
  end;
  {$IFDEF RVUNICODESTR}
  LWord := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo), RVU_GetRawUnicode(AWord));
  {$ELSE}
  LWord := AWord;
  {$ENDIF}
  if IsCyrillic then
    ConvertWinToKoi8(LWord);
  Result := aspell_speller_check(FSpeller, PRVAnsiChar(LWord), Length(LWord))=0;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.SuggestWords(const AWord: String; StyleNo: Integer;
  Suggestions: TStrings; rv: TCustomRichView);
var LWord, LSuggestedWord: TRVRawByteString;
    LSuggestedWord2: String;
    ASpellSuggestions: AspellWordList;
    Elements: AspellStringEnumeration;
    PSuggestedWord: PRVAnsiChar;
    FCyrillic: Boolean;
begin
  if FSpeller=nil then
    exit;
  {$IFDEF RVUNICODESTR}
  LWord := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo),
    RVU_GetRawUnicode(AWord));
  {$ELSE}
  LWord := AWord;
  {$ENDIF}
  FCyrillic := IsCyrillic;
  if FCyrillic then
    ConvertWinToKoi8(LWord);
  Suggestions.BeginUpdate;
  try
    Suggestions.Clear;
    ASpellSuggestions :=
      aspell_speller_suggest(FSpeller, PRVAnsiChar(LWord), Length(LWord));
    Elements := aspell_word_list_elements(ASpellSuggestions);
    repeat
      PSuggestedWord := aspell_string_enumeration_next(Elements);
      if (PSuggestedWord<>nil) then begin
        LSuggestedWord := PSuggestedWord;
        if FCyrillic then
          ConvertKoi8ToWin(LSuggestedWord);
        {$IFDEF RVUNICODESTR}
        LSuggestedWord2 := RVU_RawUnicodeToWideString(
          RVU_AnsiToUnicode(rv.RVData.GetStyleCodePage(StyleNo), LSuggestedWord));
        {$ELSE}
        LSuggestedWord2 := LSuggestedWord;
        {$ENDIF}
        Suggestions.Add(LSuggestedWord2);
      end;
    until (PSuggestedWord = nil);
    delete_aspell_string_enumeration(elements);
  finally
    Suggestions.EndUpdate;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.AddToDictionary(const AWord: String; StyleNo: Integer;
  rv: TCustomRichView);
var LWord: TRVRawByteString;
begin
  if FSpeller=nil then
    exit;
  {$IFDEF RVUNICODESTR}
  LWord := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo), RVU_GetRawUnicode(AWord));
  {$ELSE}
  LWord := AWord;
  {$ENDIF}
  if IsCyrillic then
    ConvertWinToKoi8(LWord);
  aspell_speller_add_to_personal(FSpeller, PRVAnsiChar(LWord), Length(LWord));
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.StoreReplacement(const AWord, ACorrection: String;
  StyleNo: Integer; rv: TCustomRichView);
var LWord, LCor: TRVRawByteString;
begin
  if FSpeller=nil then
    exit;
  {$IFDEF RVUNICODESTR}
  LWord := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo), RVU_GetRawUnicode(AWord));
  LCor  := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo), RVU_GetRawUnicode(ACorrection));
  {$ELSE}
  LWord := AWord;
  LCor  := ACorrection;
  {$ENDIF}
  if IsCyrillic then begin
    ConvertWinToKoi8(LWord);
    ConvertWinToKoi8(LCor);
  end;
  aspell_speller_store_replacement(FSpeller, PRVAnsiChar(LWord), Length(LWord),
    PRVAnsiChar(LCor), Length(LCor));
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.AddToIgnoreList(const AWord: String; StyleNo: Integer;
  rv: TCustomRichView);
var LWord: TRVRawByteString;
begin
  if FSpeller=nil then
    exit;
  {$IFDEF RVUNICODESTR}
  LWord := RVU_UnicodeToAnsi(rv.RVData.GetStyleCodePage(StyleNo), RVU_GetRawUnicode(AWord));
  {$ELSE}
  LWord := AWord;
  {$ENDIF}
  if IsCyrillic then
    ConvertWinToKoi8(LWord);
  aspell_speller_add_to_session(FSpeller, PRVAnsiChar(LWord), Length(LWord));
end;
{------------------------------------------------------------------------------}
procedure TRVASpell.DoSpellFormAction(mr: TModalResult; const AWord, AReplaceTo: String);
var Action: TRVASpellFormAction;
begin
  if not Assigned(FOnSpellFormAction) then
    exit;
  case mr of
    mrNo:
      Action := rvasIgnore;
    mrNoToAll:
      Action := rvasIgnoreAll;
    mrOk:
      Action := rvasAdd;
    mrYes:
      Action := rvasChange;
    mrYesToAll:
      Action := rvasChangeAll;
    else
      exit;
  end;
  FOnSpellFormAction(Self, AWord, AReplaceTo, Action);
end;
{============================= TRVASpellHelper ================================}
constructor TRVASpellHelper.Create;
begin
  inherited;
  FReplaceFrom := TStringList.Create;
  FReplaceTo   := TStringList.Create;
end;
{------------------------------------------------------------------------------}
destructor TRVASpellHelper.Destroy;
begin
  FReplaceFrom.Free;
  FReplaceTo.Free;
  frm.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVASpellHelper.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var s: String;
begin
  s := RVASpell.GetDictName;
  FCyrillic := (s='ru') or (s='uk');
  Run(Editor, Scope);
end;
{------------------------------------------------------------------------------}
function TRVASpellHelper.ProcessWord: Boolean;
var Correct, Index: Integer;
    Suggestions: AspellWordList;
    Elements: AspellStringEnumeration;
    PSuggestedWord: PRVAnsiChar;
    LWordA, LSuggestedWordA: TRVRawByteString;
    LWord, LSuggestedWord: String;
    LText: String;
    SuggestionsStrings: TStrings;
    mr: TModalResult;
    {$IFDEF RVUNICODESTR}
    CodePage: TRVCodePage;
    {$ENDIF}
    IsFormCreated: Boolean;
begin
  Result := True;
  {$IFDEF RVUNICODESTR}
  CodePage := FRVASpell.FRVData.GetItemCodePage(FItemNo);
  LWord := RVU_RawUnicodeToWideString(RVU_AnsiToUnicode(CodePage, FWordA));
  {$ELSE}
  LWord := FWordA;
  {$ENDIF}
  Index := FReplaceFrom.IndexOf(LWord);
  if Index>=0 then begin
    FWordChanged := True;
    {$IFDEF RVUNICODESTR}
    FWordA := RVU_UnicodeToAnsi(CodePage, RVU_GetRawUnicode(FReplaceTo[Index]));
    {$ELSE}
    FWordA := FReplaceTo[Index];
    {$ENDIF}
    exit;
  end;
  LWordA := FWordA;
  if FCyrillic then
    ConvertWinToKoi8(LWordA);
  Correct := aspell_speller_check(FRVASpell.FSpeller, PRVAnsiChar(LWordA), Length(LWordA));
  if Correct<>0 then
    exit;
  SelectCurrentWord;
  IsFormCreated := frm=nil;
  CreateSpellForm(TRVSpellFormStyle(FRVASpell.SpellFormStyle), frm);
  if IsFormCreated and Assigned(RVASpell.OnSpellForm) then
    RVASpell.OnSpellForm(RVASpell, frm);
  SuggestionsStrings := frm.GetStringsForSuggestions;
  SuggestionsStrings.BeginUpdate;
  SuggestionsStrings.Clear;
  Suggestions := aspell_speller_suggest(FRVASpell.FSpeller, PRVAnsiChar(LWordA), Length(LWordA));
  Elements := aspell_word_list_elements(Suggestions);
  repeat
    PSuggestedWord := aspell_string_enumeration_next(Elements);
    if (PSuggestedWord<>nil) then begin
      LSuggestedWordA := PSuggestedWord;
      if FCyrillic then
        ConvertKoi8ToWin(LSuggestedWordA);
      {$IFDEF RVUNICODESTR}
      LSuggestedWord := RVU_RawUnicodeToWideString(RVU_AnsiToUnicode(CodePage, LSuggestedWordA));
      {$ELSE}
      LSuggestedWord := LSuggestedWordA;
      {$ENDIF}
      SuggestionsStrings.Add(LSuggestedWord);
    end;
  until (PSuggestedWord = nil);
  delete_aspell_string_enumeration(elements);
  SuggestionsStrings.EndUpdate;
  if FRVASpell.SpellFormStyle=rvafMSWord then begin
    {$IFDEF RVUNICODESTR}
    LText := RVU_RawUnicodeToWideString(RVU_AnsiToUnicode(CodePage, FTextA));
    {$ELSE}
    LText := FTextA;
    {$ENDIF}
    end
  else
    LText := '';
  BeforeShowSpellForm(LWord, LText, FStartOffs, FEndOffs, frm,
    FRVData.GetRVStyle.TextStyles[FRVData.GetItemStyle(FItemNo)]);
  FDialogShown := True;
  mr := frm.ShowModal;
  LText := frm.GetCorrectedText;
  FRVASpell.DoSpellFormAction(mr, LWord, LText);
  case mr of
    mrNo: ;
    mrNoToAll:
      aspell_speller_add_to_session(FRVASpell.FSpeller, PRVAnsiChar(LWordA), Length(LWordA));
    mrOk:
      aspell_speller_add_to_personal(FRVASpell.FSpeller, PRVAnsiChar(LWordA), Length(LWordA));
    mrYes, mrYesToAll:
      begin
        LSuggestedWord := LText;
        if mr=mrYesToAll then begin
          FReplaceFrom.Add(LWord);
          FReplaceTo.Add(LSuggestedWord);
        end;
        {$IFDEF RVUNICODESTR}
        LSuggestedWordA := RVU_UnicodeToAnsi(CodePage, RVU_GetRawUnicode(LSuggestedWord));
        {$ELSE}
        LSuggestedWordA := LSuggestedWord;
        {$ENDIF}
        if FCyrillic then
          ConvertWinToKoi8(LSuggestedWordA);
        if not frm.IsReplacingItemText then begin
          aspell_speller_store_replacement(FRVASpell.FSpeller,PRVAnsiChar(LWordA), Length(LWordA),
            PRVAnsiChar(LSuggestedWordA), Length(LSuggestedWordA));
          FWordA := LSuggestedWordA;
          FWordChanged := True;
          end
        else begin
          FTextA := LSuggestedWordA;
          FTextChanged := True;
        end;
      end;
    mrCancel:
      Result := False;
  end;
end;
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVASpell]);
end;

end.