
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVVSSpell8: component for                      }
{       spell-checking RichViewEdit using VSSpell 8.0   }
{       by ComponentOne LLC                             }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       VSSpell : Copyright (c) ComponentOne LLC        }
{       http://www.componentone.com                     }
{                                                       }
{*******************************************************}

{
  VSSpell can be downloaded from http://www.componentone.com
  This is a shareware spell-checker (ActiveX)

  Installing VSSpell 8.0 in Delphi IDE:
    - Menu: Component | Import ActiveX Control
    - Choose "ComponentOne Spell 8.0 Control (Version 1.0)"
    - Since its name conflicts with VSSpell 3.0 by Visual Components, Inc
      (shipped with Delphi, see ActiveX page),
      modify the name of control in "Class names" field from 'TVSSpell' to
      'TVSSpell8'.
      (If you installed it under the original name, uncomment the line
       // TVSSpell8 = TVSSpell;
       below)
    - Click "Install" button and install the resulting file (VSSPELL8Lib_TLB)
      in some package. Compile this package.

  Installing this component:
    - Add this file (RVCOVSSpell) in some package. Compile this package.
      TRVVSSpell8 component will appear on "RichView" page of the
      Component Palette. This component should be used instead of TVSSpell8.

  Using:
    - place TVSSpell8 on the form.
    - in program, call TVSSpell8.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call TVSSpell8.Check(rve, rvesFromStart) to check from the
      beginning.

  Functions for live spelling:
    - the ActiveX control must be initialized before starting live spelling.
      Call RVVSSpell8.BadWordDialog at the beginning of your application
      (for example, in Form.OnCreate). There may be more elegant solutions,
      but it works
    - checking word in RichViewEdit.OnSpellingCheck: TRVVSSpell8.IsMisspelled
    - adding to ignore list: TRVVSSpell8.AddIgnoreAll := word
    - adding to dictionary: TRVVSSpell8.AddCustomWord := word
    - suggestions: TRVVSSpell8.SuggestWords
    - event on adding word to ignore list: TRVVSSpell8.OnDialogAction,
       parameter Action=4 (*DIALOGACTION_IGNOREALL*), the word is
       TRVVSSpell8.CheckWord
    - event on adding word to dictionary: the same,
       parameter Action=5 (*DIALOGACTION_ADD*),

  History:
  2011-03-20: property DialogShown specifies if a spelling check dialog has been
    shown during the last spelling check.
  2004-08-02: methods for live spelling check       
}


unit RVCOVSSpell;

interface
uses SysUtils, Windows, Classes, Forms,
     RVEdit,
     RVWordEnum,
     VSSPELL8Lib_TLB;

type

  // TVSSpell8 = TVSSpell;

  TRVVSSpell8BadWord = procedure(Sender: TObject; var Problem: Smallint) of object;
  TRVVSSpell8Complete = procedure(Sender: TObject; var Cancelled: Smallint) of object;

  TRVVSSpell8 = class;

  TRVVSSpell8Helper = class (TRVWordEnumerator)
    private
      FOldOnComplete: TRVVSSpell8Complete;
      FOldOnBadWord: TRVVSSpell8BadWord;
      FOldOnChanged: TNotifyEvent;
      FCancelled: Boolean;
      FVSSpell: TRVVSSpell8;
      FDialogShown: Boolean;
      procedure DoOnBadWord(Sender: TObject; var Problem: Smallint);
      procedure DoOnComplete(Sender: TObject; var Cancelled: Smallint);
      procedure DoOnChanged(Sender: TObject);
    protected
      function ProcessWord: Boolean; override;
    public
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      property DialogShown: Boolean read FDialogShown;
    published
      property VSSpell: TRVVSSpell8 read FVSSpell write FVSSpell;
  end;

  TRVVSSpell8 = class (TVSSpell8)
    private
      FComplete: Boolean;
      FMisspelled: Boolean;
      procedure DoOnBadWord(Sender: TObject; var Problem: Smallint);
      procedure DoOnComplete(Sender: TObject; var Cancelled: Smallint);
    public
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      function IsMisspelled(const AWord: String): Boolean;
      procedure SuggestWords(const AWord: String; ASuggestions: TStrings);
  end;

procedure Register;

implementation

{=============================== TRVVSSpell8Helper ============================}
procedure TRVVSSpell8Helper.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
begin
  FCancelled := False;
  FOldOnComplete := FVSSpell.OnComplete;
  FOldOnBadWord  := FVSSpell.OnBadWord;
  FOldOnChanged  := FVSSpell.OnChanged;
  FVSSpell.OnComplete := DoOnComplete;
  FVSSpell.OnBadWord  := DoOnBadWord;
  FVSSpell.OnChanged  := DoOnChanged;
  try
    Run(Editor, Scope);
  finally
    FVSSpell.OnComplete := FOldOnComplete;
    FVSSpell.OnBadWord  := FOldOnBadWord;
    FVSSpell.OnChanged  := FOldOnChanged;
    FOldOnComplete := nil;
    FOldOnBadWord  := nil;
    FOldOnChanged  := nil;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8Helper.DoOnBadWord(Sender: TObject; var Problem: Smallint);
begin
  SelectCurrentWord;
  FDialogShown := True;
  if Assigned(FOldOnBadWord) then
    FOldOnBadWord(Sender, Problem);
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8Helper.DoOnChanged(Sender: TObject);
begin
  FWordChanged := True;
  FWordA := FVSSpell.Text;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8Helper.DoOnComplete(Sender: TObject;
  var Cancelled: Smallint);
begin
  if Assigned(FOldOnComplete)  then
    FOldOnComplete(Sender, Cancelled);
  FCancelled := Cancelled<>0;
  FVSSpell.FComplete := True;
end;
{------------------------------------------------------------------------------}
function TRVVSSpell8Helper.ProcessWord: Boolean;
begin
  if (Length(FWordA) > 49) then  // VSSpell fails if word greater than 49 characters
  begin
    Result := True;
    exit;
  end;
  FVSSpell.Text := FWordA;
  FVSSpell.FComplete := False;
  FVSSpell.CheckText;
  while not FVSSpell.FComplete do
    Application.ProcessMessages;
  Result := not FCancelled;
end;
{=============================== TRVVSSpell8 ==================================}
procedure TRVVSSpell8.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var Helper: TRVVSSpell8Helper;
begin
  Helper := TRVVSSpell8Helper.Create;
  try
    Helper.VSSpell := Self;
    Helper.Check(Editor, Scope);
    FDialogShown := Helper.DialogShown;
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8.DoOnBadWord(Sender: TObject; var Problem: Smallint);
begin
  FMisspelled := True;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8.DoOnComplete(Sender: TObject;
  var Cancelled: Smallint);
begin
  FComplete := True;
end;
{------------------------------------------------------------------------------}
function TRVVSSpell8.IsMisspelled(const AWord: String): Boolean;
var FOldOnBadWord: TRVVSSpell8BadWord;
    FOldOnComplete: TRVVSSpell8Complete;
    FOldBadWordDialog: BadWordDialogConstants;
begin
  if (Length(AWord) > 49) then  // VSSpell fails if word greater than 49 characters
  begin
    Result := False;
    exit;
  end;
  FOldOnBadWord := OnBadWord;
  FOldOnComplete := OnComplete;
  FOldBadWordDialog := BadWordDialog;
  try
    FMisspelled := False;
    BadWordDialog := vsspellNoDialog;
    OnBadWord := DoOnBadWord;
    OnComplete := DoOnComplete;
    FComplete := False;
    CheckWord := AWord;
    while not FComplete do;
  finally
    OnBadWord := FOldOnBadWord;
    OnComplete := FOldOnComplete;
    BadWordDialog := FOldBadWordDialog;
  end;
  Result := FMisspelled;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell8.SuggestWords(const AWord: String;
  ASuggestions: TStrings);
var i: Integer;
begin
  ASuggestions.BeginUpdate;
  try
    ASuggestions.Clear;
    if IsMisspelled(AWord) then
      for i := 0 to SuggestionCount-1 do
        ASuggestions.Add(Suggestion[i]);
  finally
    ASuggestions.EndUpdate;
  end;
end;
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVVSSpell8]);
end;


end.