
{*******************************************************}
{                                                       }
{       TRvDxSpellChecker - component for checking      }
{       spelling in TRichViewEdit controls using        }
{       ExpressSpellChecker                             }
{                                                       }
{       TRichView (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       ExpressSpellChecker (c) Developer Express Inc.  }
{       http://www.devexpress.com                       }
{                                                       }
{*******************************************************}

{$I RV_Defs.inc}

{$DEFINE RVUSECXTRICHVIEW} // remove if you do not use TcxTRichViewEdit

{

  This component allows using ExpressSpellChecker with the following components:
  - TRichViewEdit, TDBRichViewEdit
  - ScaleRichView (TSRichViewEdit, TDBSRichViewEdit)
  - TcxTRichViewEdit, TcxDBTRichViewEdit

  Installing this component:

    - if you do not use TcxTRichViewEdit and TcxDBTRichViewEdit, comment the
      declatation RVUSECXTRICHVIEW above. These components are available at
      http://www.trichview.com/resources/devexpress/

    - Add this file (RVDXSpell.pas) in some package. Compile this package.
      TRvDxSpellChecker component will appear on "RichView" page of the
      Component Palette.

  Using:
    - place TRVDXSpell on the form.
    - in program, call TRVDXSpell.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call TRVDXSpell.Check(rve, rvesFromStart) to check from the
      beginning.
      (if you use TSRichViewEdit, call Check(srve.RichViewEdit)
      (if you use TcxTRichViewEdit, call Check(cxrve.InnerEditor)


  Live spelling in TcxTRichViewEdit: works automatically, does not require
    any code.

  Functions for live spelling in TRichViewEdit and TSRichViewEdit:
    - (IMPORTANT!) before starting live spelling, call TRVDXSpell.UpdateRules
    - checking word in RichViewEdit.OnSpellingCheck: TRVDXSpell.IsMisspelled
    - adding to ignore list: TRVDXSpell.IgnoreList.Add
    - adding to replace-all list: TRVDXSpell.ChangeList.Add
    - adding to dictionary: TRVDXSpell.AddWordToUserDictionary
    - suggestions: TRVDXSpell.SuggestWords or GetSuggestions
    - event on adding word to ignore list: TRVDXSpell.OnSpellFormAction
    - event on adding word to dictionary: TRVDXSpell.OnSpellFormAction

}

unit RVDXSpell;

interface
uses  Windows, SysUtils, Classes, Controls, Forms, Menus,
      dxSpellCheckerCore,
      dxSpellChecker, RichView, RVERVData,
      RVEdit, RVWordEnum, CRVData, RVUni,
      {$IFDEF RVUSECXTRICHVIEW}
      dxSpellCheckerCheckAsYouType, dxSpellCheckerAutoCorrect,
      dxSpellCheckerRules, cxClasses, cxGeometry,
      dxCore, dxSpellCheckerStrs, cxTRichView,
      {$ENDIF}
      RVTypes;

type
  TRvDxSpellChecker = class;
  TRVdxSpellHelper = class;

  TRVdxSpellFormAction = (rvdxIgnore, rvdxIgnoreAll, rvdxChange, rvdxChangeAll,
    rvdxAdd);
  TRVDXSpellFormActionEvent = procedure(Sender: TRvDxSpellChecker;
    const AWord, AReplaceTo: TRVUnicodeString; Action: TRVdxSpellFormAction;
    AReplacingItemText: Boolean) of object;

  {$IFDEF RVUSECXTRICHVIEW}
  TRvDxSpellCheckerCheckAsYouTypeManager = class(TdxSpellCheckerCheckAsYouTypeManager)
    private
      FFinalizing: Boolean;
      FStoredCheckAsYouTypeOptions: Boolean;
      FOldOnSpellingCheck: TRVSpellingCheckEvent;
      FEditor: TCustomRichViewEdit;
    protected
      procedure FinalizeController; override;
      procedure InitializeController(AControl: TWinControl); override;
      function CanPopup(const P: TPoint): Boolean; override;
      procedure CreateItems(APopup: TPopupMenu; AOwner: TComponent); override;
      procedure RVMenuItemClickHandler(ASender: TObject);
    public
      procedure CheckFinish; override;
      procedure CheckStart(AControl: TWinControl); override;
      function QueryPopup(APopup: TComponent; const P: TPoint): Boolean; override;
    end;

  TRvdxSpellCheckerAutoCorrectManager = class(TdxSpellCheckerAutoCorrectManager)
    protected
      procedure InitializeController(AControl: TWinControl); override;
  end;
  {$ENDIF}

  {$IFDEF RICHVIEWDEFXE2}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
  {$ENDIF}

  TRvDxSpellChecker = class (TdxSpellChecker)
    private
      FOnSpellFormAction: TRVDXSpellFormActionEvent;
      FDummyParser: TdxSpellCheckerSingleWordParser;
      FDialogShown: Boolean;
      procedure DoSpellFormAction(mr: TModalResult; const AWord, AReplaceTo: String;
        AReplacingItemText: Boolean);
    {$IFDEF RVUSECXTRICHVIEW}
      procedure DoRVSpellingCheck(Sender: TCustomRichView; const AWord: String;
        StyleNo: Integer; var Misspelled: Boolean);
    protected
      function CreateCheckAsYouTypeManager: TdxSpellCheckerCustomCheckAsYouTypeManager; override;
      function CreateAutoCorrectManager: TdxSpellCheckerCustomAutoCorrectManager; override;
    {$ENDIF}
    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      function IsMisspelled(const AWord: TRVUnicodeString): Boolean;
      procedure  SuggestWords(const AWord: TRVUnicodeString;
        Suggestions: TStrings);
      procedure UpdateRules;
      procedure Check(AEdit: TCustomRichViewEdit;
        Scope: TRVEnumScope = rvesFromCursor); overload;
      property IgnoreList;
      property DialogShown: Boolean read FDialogShown;
    published
      property OnSpellFormAction: TRVDXSpellFormActionEvent
        read FOnSpellFormAction write FOnSpellFormAction;
  end;

  TRVdxSpellHelper = class (TRVWordEnumerator)
    private
      FChecker: TRvDxSpellChecker;
      FDialog: TForm;
      procedure CreateSpellDialog;
    protected
      function ProcessWord: Boolean; override;
    public
      constructor Create;
      destructor Destroy; override;
  end;

  procedure Register;

implementation
uses RVDXSpellCustomForm, RVDXSpellOutlookForm, RVDXSpellWordForm;

{==================================== TRVASpell ===============================}
procedure TRvDxSpellChecker.Check(AEdit: TCustomRichViewEdit; Scope: TRVEnumScope);
var Helper: TRVdxSpellHelper;
begin
  CheckAsYouTypeManager.BeginManualCheck;
  try
    UpdateRules;
    Helper := TRVdxSpellHelper.Create;
    try
      Helper.FChecker := Self;
      Helper.Run(AEdit, Scope);
      FDialogShown := Helper.DialogShown;
    finally
      Helper.Free;
    end;
  finally
    CheckAsYouTypeManager.EndManualCheck;
  end;
end;
{------------------------------------------------------------------------------}
constructor TRvDxSpellChecker.Create(AOwner: TComponent);
begin
  inherited;
  FDummyParser := TdxSpellCheckerSingleWordParser.Create(Self);
end;
{------------------------------------------------------------------------------}
{$IFDEF RVUSECXTRICHVIEW}
function TRvDxSpellChecker.CreateCheckAsYouTypeManager: TdxSpellCheckerCustomCheckAsYouTypeManager;
begin
  Result := TRvdxSpellCheckerCheckAsYouTypeManager.Create(Self);
end;
{------------------------------------------------------------------------------}
function TRvDxSpellChecker.CreateAutoCorrectManager: TdxSpellCheckerCustomAutoCorrectManager;
begin
  Result := TRvdxSpellCheckerAutoCorrectManager.Create(Self);
end;
{$ENDIF}
{------------------------------------------------------------------------------}
destructor TRvDxSpellChecker.Destroy;
begin
  FDummyParser.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
type
  TdxSpellCheckerSpellingOptionsHack = class (TdxSpellCheckerSpellingOptions)
  end;

procedure TRvDxSpellChecker.UpdateRules;
begin
  CheckingLock.Enter;
  try
    Rules.Clear;
    TdxSpellCheckerSpellingOptionsHack(SpellingOptions).PopulateRules(Rules, FDummyParser);
  finally
    CheckingLock.Leave;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellChecker.DoSpellFormAction(mr: TModalResult;
  const AWord, AReplaceTo: String; AReplacingItemText: Boolean);
var Action: TRVDXSpellFormAction;
begin
  if not Assigned(FOnSpellFormAction) then
    exit;
  case mr of
    mrNo:
      Action := rvdxIgnore;
    mrNoToAll:
      Action := rvdxIgnoreAll;
    mrOk:
      Action := rvdxAdd;
    mrYes:
      Action := rvdxChange;
    mrYesToAll:
      Action := rvdxChangeAll;
    else
      exit;
  end;
  FOnSpellFormAction(Self, AWord, AReplaceTo, Action, AReplacingItemText);
end;
{------------------------------------------------------------------------------}
function TRvDxSpellChecker.IsMisspelled(const AWord: TRVUnicodeString): Boolean;
begin
  Result := (Length(AWord)>1) and not IsValidWord(AWord) and
    not IgnoreList.Find(AWord);
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellChecker.SuggestWords(const AWord: TRVUnicodeString;
  Suggestions: TStrings);
var DXSuggestions: TdxSpellCheckerSuggestionList;
begin
  DXSuggestions := GetSuggestions(AWord);
  try
    Suggestions.Clear;
    DXSuggestions.SaveToStrings(Suggestions)
  finally
    DXSuggestions.Free;
  end;
end;
{------------------------------------------------------------------------------}
{$IFDEF RVUSECXTRICHVIEW}
procedure TRvDxSpellChecker.DoRVSpellingCheck(Sender: TCustomRichView;
  const AWord: String; StyleNo: Integer; var Misspelled: Boolean);
begin
  Misspelled := IsMisspelled(AWord);
end;
{$ENDIF}
{=============================== TRVdxSpellHelper =============================}
constructor TRVdxSpellHelper.Create;
begin
  inherited Create;
  Unicode := True;
end;
{------------------------------------------------------------------------------}
destructor TRVdxSpellHelper.Destroy;
begin
  FDialog.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVdxSpellHelper.CreateSpellDialog;
begin
  if FDialog<>nil then
    exit;
  case FChecker.SpellingFormType of
   sftWord:
     FDialog := TfrmRVDXSpellCheckerWordForm.Create(Application);
   else
     FDialog := TfrmRVDXSpellCheckerOutlookForm.Create(Application);
  end;
end;
{------------------------------------------------------------------------------}
function TRVdxSpellHelper.ProcessWord: Boolean;
var
  RepItem: TdxSpellCheckerReplacement;
var mr: TModalResult;
    NewText: TRVUnicodeString;
begin
  if FChecker.DeleteList.Find(FWordW) then begin
    SelectCurrentWord;
    FWordW := '';
    FWordChanged := True;
    Result := True;
    exit;
  end;
  RepItem := FChecker.ChangeList.FindReplacement(FWordW);
  if RepItem<>nil then begin
    SelectCurrentWord;
    FWordW := RepItem.Replacement;
    FWordChanged := True;
    Result := True;
    exit;
  end;
  Result := not FChecker.IsMisspelled(FWordW);
  if Result then
    exit;
  Result := True;
  CreateSpellDialog;
  SelectCurrentWord;
  if FDialog is TfrmRVDXSpellCheckerWordForm then
    TfrmRVDXSpellCheckerWordForm(FDialog).Init(FChecker, FWordW, FTextW,
      FStartOffs-1, FEndOffs-FStartOffs,
      FRVData.GetRVStyle.TextStyles[FRVData.GetItemStyle(FItemNo)])
  else if FDialog is TfrmRVDXSpellCheckerOutlookForm then
    TfrmRVDXSpellCheckerOutlookForm(FDialog).Init(FChecker, FWordW,
      FRVData.GetRVStyle.TextStyles[FRVData.GetItemStyle(FItemNo)]);
  FDialogShown := True;
  mr := FDialog.ShowModal;
  NewText := TfrmRVDXSpellCheckerCustomForm(FDialog).GetCorrectedText;
  FChecker.DoSpellFormAction(mr, FWordW, NewText,
    TfrmRVDXSpellCheckerCustomForm(FDialog).IsReplacingItemText);
  case mr of
    mrNoToAll: // ignore all
      FChecker.IgnoreList.Add(FWordW);
    mrYes: // replace/delete
      begin
        if TfrmRVDXSpellCheckerCustomForm(FDialog).IsReplacingItemText then begin
          FTextChanged := True;
          FTextW := NewText;
          end
        else begin
          FWordChanged := True;
          FWordW := NewText;
        end;
      end;
    mrYesToAll: // replace/delete all
      begin
        if NewText='' then
          FChecker.DeleteList.Add(FWordW)
        else
          FChecker.ChangeList.Add(FWordW, NewText);
        FWordW := NewText;
        FWordChanged := True;
      end;
    mrOk: // add
      FChecker.AddWordToUserDictionary(FWordW);
    mrCancel:
      Result := False;
  end;
end;
{================ TRvDxSpellCheckerCustomCheckAsYouTypeManager ================}
{$IFDEF RVUSECXTRICHVIEW}
function TRvDxSpellCheckerCheckAsYouTypeManager.CanPopup(
  const P: TPoint): Boolean;
var Wrd: String;
    StyleNo: Integer;
    AItem: TdxSpellCheckerMisspelledWord;
    pt: TPoint;
    R: TRect;
begin
  if FEditor<>nil then begin
    Result := (Options.PopupMenuItems <> []) and
      FEditor.FocusedEx and FEditor.GetCurrentMisspelling(False, Wrd, StyleNo);
    if Result then begin
      AItem := TdxSpellCheckerMisspelledWord.Create(Wrd, nil, nil);
      GetCaretPos(pt);
      R.TopLeft := pt;
      R.BottomRight := pt;
      inc(R.Bottom, FEditor.TopLevelEditor.RVData.DrawItems[
        TRVEditRVData(FEditor.TopLevelEditor.RVData).CaretDrawItemNo].Height);
      AItem.AddRect(R);
      MisspelledItem := AItem;
      AItem.Free;
      end
    else
      MisspelledItem := nil;
    end
  else
    Result := inherited CanPopup(P);
end;
{------------------------------------------------------------------------------}
function TRvDxSpellCheckerCheckAsYouTypeManager.QueryPopup(APopup: TComponent;
  const P: TPoint): Boolean;
  function GetPopupPos: TPoint;
  var
    R: TRect;
  begin
    with MisspelledItem do
      R := UnderlineRect[UnderlineRectCount - 1];
    Result := cxPoint(R.Left, R.Bottom);
    if not PtInRect(FEditor.ClientRect, cxPoint(R.Left, R.Top)) then
    begin
      GetCaretPos(Result);
      Result.Y := R.Bottom;
    end;
    Result := FEditor.ClientToScreen(Result);
  end;

var
  AItemsOwner: TComponent;
begin
  if FEditor=nil then begin
    Result := inherited QueryPopup(APopup, P);
    exit;
  end;
  Result := CanPopup(P);
  if Result then
  begin
    if Options.PopupMenu <> nil then
    begin
      InnerShowPopupMenu(Options.PopupMenu, GetPopupPos);
      Exit;
    end;
    AItemsOwner := TComponent.Create(nil);
    try
      if not Options.ModifyControlPopupMenu or not (APopup is TPopupMenu) then
      begin
        CreateInternalPopupMenu;
        APopup := InternalPopupMenu;
      end;
      CreateItems(TPopupMenu(APopup), AItemsOwner);
      InnerShowPopupMenu(APopup, GetPopupPos);
      Application.ProcessMessages;
    finally
      AItemsOwner.Free;
    end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellCheckerCheckAsYouTypeManager.CheckFinish;
begin
  if FEditor<>nil then
    FinalizeController;
  inherited CheckFinish;
end;
procedure TRvDxSpellCheckerCheckAsYouTypeManager.CheckStart(
  AControl: TWinControl);
begin
  if Active and (AControl is TcxInnerTRichViewEdit) then
    Edit := AControl
  else
    inherited;
end;
{------------------------------------------------------------------------------}
function GetMisspelledWord(rve: TCustomRichViewEdit): TRVUnicodeString;
var StyleNo: Integer;
  WordA: String;
begin
  rve.GetCurrentMisspelling(False, WordA, StyleNo);
  {$IFDEF RVUNICODESTR}
  Result := WordA;
  {$ELSE}
  Result := RVU_RawUnicodeToWideString(
    RVU_AnsiToUnicode(rve.RVData.GetStyleCodePage(StyleNo), WordA));
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
procedure SelectMisspelledWord(rve: TCustomRichViewEdit);
var StyleNo: Integer;
    AWord: String;
begin
  rve.GetCurrentMisspelling(True, AWord, StyleNo);
end;
{------------------------------------------------------------------------------}
procedure IgnoreMisspeledWord(rve: TCustomRichViewEdit);
var StyleNo: Integer;
    AWord: String;
begin
  rve.GetCurrentMisspelling(False, AWord, StyleNo);
  rve.LiveSpellingValidateWord(AWord);
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellCheckerCheckAsYouTypeManager.RVMenuItemClickHandler(ASender: TObject);
  procedure SuggestionClick(AIndex: Integer; AIsAutoCorrect: Boolean);
  var
    ASuggestionList: TdxSpellCheckerSuggestionList;
  begin
    ASuggestionList := SpellChecker.GetSuggestions(MisspelledItem.Word);
    try
      if ASuggestionList.Count >= AIndex then begin
        SelectMisspelledWord(FEditor);
        FEditor.InsertTextW(ASuggestionList[AIndex].Word);
      end;
    finally
      FreeAndNil(ASuggestionList);
    end;
  end;

var
  ACommand: Integer;
  AIndex: Integer;
begin
  if Assigned(MisspelledItem) then
  begin
    ACommand := TMenuItem(ASender).Tag;
    case ACommand of
      Ord(scmiIgnoreAll):
        begin
          TRvDxSpellChecker(SpellChecker).IgnoreList.Add(MisspelledItem.Word);
          IgnoreMisspeledWord(FEditor);
        end;
      Ord(scmiAddToDictionary):
        begin
          TRvDxSpellChecker(SpellChecker).AddWordToUserDictionary(MisspelledItem.Word);
          IgnoreMisspeledWord(FEditor);
        end;
      Ord(scmiDelete):
        begin
          SelectMisspelledWord(FEditor);
          FEditor.DeleteSelection;
        end;
      Ord(scmiSpelling):
        TRvDxSpellChecker(SpellChecker).Check(FEditor);
    else
      if (ACommand >= Ord(scmiSuggestions)) or (ACommand < -1) then
      begin
        if ACommand < -1 then
          AIndex := -ACommand - 2
        else
          AIndex := ACommand - Ord(scmiSuggestions);
        SuggestionClick(AIndex, ACommand < -1);
      end;
    end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellCheckerCheckAsYouTypeManager.CreateItems(APopup: TPopupMenu;
  AOwner: TComponent);
var WordW: TRVUnicodeString;
    I: Integer;

  procedure AddItem(var APosition: Integer; ATag: Integer; const ACaption: WideString;
    ASubMenuItem: TMenuItem = nil);
  var
    AItem: TMenuItem;
  begin
    AItem := TMenuItem.Create(AOwner);
    AItem.Caption := String(ACaption);
    if not AItem.IsLine then
    begin
      AItem.Tag := ATag;
      AItem.OnClick := RVMenuItemClickHandler;
      AItem.AutoHotkeys := maManual;
    end;
    if ASubMenuItem = nil then
      APopup.Items.Insert(APosition, AItem)
    else
      ASubMenuItem.Insert(ASubMenuItem.Count, AItem);
    Inc(APosition);
  end;
  {
  procedure GenerateAutoCorrectItems(var APosition: Integer; ASuggestionList: TdxSpellCheckerSuggestionList);
  var
    I: Integer;
    ASubMenuItem: TMenuItem;
  begin
    if APosition > 0 then
      AddItem(APosition, 0, cLineCaption);
    AddItem(APosition, -1, cxGetResourceString(@sdxSpellCheckerAutoCorrect));
    ASubMenuItem := AOwner.Components[AOwner.ComponentCount - 1] as TMenuItem;
    I := 0;
    if ASuggestionList.Count > 0 then
    begin
      while (I < ASuggestionList.Count) and (I < Options.SuggestionCount) do
        AddItem(I, -I - 2, ASuggestionList[I].Word, ASubMenuItem);
      AddItem(I, 0, cLineCaption, ASubMenuItem);
    end;
    AddItem(I, Ord(scmiAutoCorrect), cxGetResourceString(@sdxSpellCheckerAutoCorrectOptionsFormCaption), ASubMenuItem);
  end;
  }
  procedure GenerateItemsForMisspelledWord(var APosition: Integer);
  var
    ASuggestionList: TdxSpellCheckerSuggestionList;
  begin
    ASuggestionList := SpellChecker.GetSuggestions(WordW);
    try
      APosition := 0;
      if scmiSuggestions in Options.PopupMenuItems then
        while (APosition < ASuggestionList.Count) and (APosition < Options.SuggestionCount) do
          AddItem(APosition, Ord(High(TdxSpellCheckerPopupMenuItem)) + 1 + APosition, ASuggestionList[APosition].Word);
      if APosition > 0 then
        AddItem(APosition, 0, cLineCaption);
      if scmiIgnoreAll in Options.PopupMenuItems then
        AddItem(APosition, Ord(scmiIgnoreAll), cxGetResourceString(@sdxSpellCheckerIgnoreAllButton));
      if (scmiAddToDictionary in Options.PopupMenuItems) and SpellChecker.HasEnabledUserDictionary then
        AddItem(APosition, Ord(scmiAddToDictionary), cxGetResourceString(@sdxSpellCheckerAddButton));
      {
      if SpellChecker.AutoCorrectOptions.Active and (scmiAutoCorrect in Options.PopupMenuItems) then
        GenerateAutoCorrectItems(APosition, ASuggestionList);
      }
    finally
      FreeAndNil(ASuggestionList);
    end;
  end;

begin
  if FEditor=nil then begin
    inherited;
    exit;
  end;
  WordW := GetMisspelledWord(FEditor);
  I := 0;
  GenerateItemsForMisspelledWord(I);
  if I > 0 then
    AddItem(I, 0, cLineCaption);
  if scmiSpelling in Options.PopupMenuItems then
  begin
    AddItem(I, Ord(scmiSpelling), cxGetResourceString(@sdxSpellCheckerSpellingFormCaption));
    AddItem(I, 0, cLineCaption);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellCheckerCheckAsYouTypeManager.FinalizeController;
begin
  if FFinalizing then
    exit;
  inherited;
  if FEditor<>nil then begin
    FEditor.ClearLiveSpellingResults;
    FEditor.OnSpellingCheck := FOldOnSpellingCheck;
    FEditor := nil;
    FFinalizing := True;
    try
      SpellChecker.CheckAsYouTypeOptions.Active := FStoredCheckAsYouTypeOptions;
    finally
      FFinalizing := False;
    end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRvDxSpellCheckerCheckAsYouTypeManager.InitializeController(
  AControl: TWinControl);
begin
  if AControl is TCustomRichViewEdit then begin
    FStoredCheckAsYouTypeOptions := SpellChecker.CheckAsYouTypeOptions.Active;
    SpellChecker.CheckAsYouTypeOptions.Active := False;
    FEditor := TCustomRichViewEdit(AControl).GetRootEditor;
    //CreateAdapter(TdxSpellCheckerCustomEditAdapter, FEditor);
    (SpellChecker as TRvDxSpellChecker).UpdateRules;
    FOldOnSpellingCheck := FEditor.OnSpellingCheck;
    FEditor.OnSpellingCheck :=
      (SpellChecker as TRvDxSpellChecker).DoRVSpellingCheck;
    FEditor.StartLiveSpelling;
    end
  else
    inherited;
end;
{========================= TRvdxSpellCheckerAutoCorrectManager ================}
procedure TRvdxSpellCheckerAutoCorrectManager.InitializeController(
  AControl: TWinControl);
begin
  if not (AControl is TCustomRichViewEdit) then
    inherited;
end;
{$ENDIF}
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRvDxSpellChecker]);
end;


end.