
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVHunSpell: component for                      }
{       spell-checking RichViewEdit using               }
{       HunSpell                                        }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       HunSpell : Copyright (c) Laszlo Nemeth          }
{       http://hunspell.sourceforge.net/                }
{                                                       }
{       The first version of this component was         }
{       created by Matvey Karpushin                     }
{       matvey@e1.ru                                    }
{                                                       }
{*******************************************************}

{.$DEFINE USERVTNT}

{

  This component is a wrapper for HunSpell DLL (hunspelldll.dll),
  License: MPL/GPL/LGPL.


  Installing this component:

    - Add this file (RVHunSpell.pas) in some package. Compile this package.
      TRVHunSpell component will appear on "RichView" page of the
      Component Palette.

  Using:
    - place TRVHunSpell on the form.
    - specify DictDir property (a path to dictionary files; you can leave
      it empty if dictionaries are in the application's directory
    - add the necessary dicionaries in Dictionaries collection:
      - RVHunSpell1.Dictionaries.Add.Load(<language>), where <language> is
        a dictionary name (for example, for en_US.dic it is 'en_US';
        On successfull loading, Language property of the new collection item
        is <language>; To load custom Dictionary, call LoadCustomDic.
      - or call LoadAllDictionaries to load all dictionaries in the path and
        custom dictionary.
    - in program, call TRVHunSpell.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call TRVHunSpell.Check(rve, rvesFromStart) to check from the
      beginning.
  Custom dictionary:
    Custom dictionary is a plain text Unicode (UTF-16) file. It can be edited
    in any Unicode text editor (for example, Notepad). If you want to edit this
    file in your applications, call SaveCustomDic before editing and
    LoadCustomDic after.
  Properties:
    - DictDir: String - path to dictionaries;
    - CustomDictionaryFileName - file name of custom dictionaries,
      'customdic.txt' by default. If it does not contain a full path, it is
      searched in the application's path.
    - Dictionaries: TDictionaries - list of dictionaries; *all* these
      dictionaries are used.
    - SpellFormStyle defines which dialog style is used
      (classic or like in MS Word)
  Events:
    - OnSpellForm - event occuring when spell form is created. Can be used for
      localization.
    - OnSpellFormAction - event occuring when the user pressed a button in the
      spell form.
  This component does not support custom dictionaries.
  Functions for live spelling:
    - checking word in RichViewEdit.OnSpellingCheck: TRHunSpell.Spell
    - adding to ignore list: TRVHunSpell.AddToIgnoreList
    - adding to custom dictionary: TRVHunSpell.AddToDictionary
    - suggestions: TRVHunSpell.Suggest or SuggestW
    - event on adding word to ignore list: TRVHunSpell.OnSpellFormAction

  History:
  2014-08-08: complete Unicode support in non-Unicode versions of Delphi using
    TNT Controls (TMS Unicode Component Pack - http://www.tmssoftware.com/site/tmsuni.asp).
    To enable, remove the dot from the line .$DEFINE USERVTNT above
  2014-01-25: 64-bit compatibility
  2012-03-22: some unnecessary procedures were deleted; a newer version of hunspelldll.dll
    is supported.
  2011-03-20: property DialogShown specifies if a spelling check dialog has been
    shown during the last spelling check.
  2010-09-09:
    new: DllDir property. If not empty, specifies path (without file name) to
      hunspelldll.dll. If empty, it is searched in the application directory
      and in %PATH%
  2009-03-08:
    new: custom dictionary
  2009-02-28:
    new: the SuggestW method for D2005+ fills TWideStringList
    impr: new dictionary encodings are supported (ISO 8859), improvements in
      KOI-8 and ANSI support
    impr: the Suggest method does not add duplicates any more
    impr: for D2005-D2007, TWideStringList type is used for ignore and
      replace-all lists
  2009-02-03:
    new: support for MS Word-like spelling dialog (SpellFormStyle property)
}

unit RVHunSpell;

interface

{$I RV_Defs.inc}

uses Classes, Windows, SysUtils, Controls, Forms,
    {$IFDEF USERVTNT}TntClasses,{$ENDIF}
    {$IFDEF RICHVIEWDEF2005}WideStrings,{$ENDIF}
    RVUni, RVTypes, RVWordEnum,
    {$IFDEF USERVTNT}
    RVTntCustomSpellFrm,
    {$ELSE}
    RVCustomSpellFrm,
    {$ENDIF}
    RVEdit, RVStyle;

const
  HUNSPELLDLL = 'hunspelldll.dll';
  DEFCUSTOMDICNAME = 'customdic.txt';


type
  TDictionary = class;
  TDictionaries = class;
  TRVHunSpell = class;

  {$IFDEF RICHVIEWDEF2005}
  {$IFNDEF RICHVIEWDEF2009}
    {$DEFINE USEWIDESTRINGLISTS}
  {$ENDIF}
  {$ENDIF}

  {$IFDEF USERVTNT}
  TRVHunSpellStringList = TTntStringList;
  TRVHunSpellCustomSpell = TfrmRVTntCustomSpell;
  {$ELSE}
  TRVHunSpellCustomSpell = TfrmRVCustomSpell;
  {$IFDEF USEWIDESTRINGLISTS}
  TRVHunSpellStringList = TWideStringList;
  {$ELSE}
  TRVHunSpellStringList = TStringList;
  {$ENDIF}
  {$ENDIF}
  


  TDictionary = class(TCollectionItem)
  private
    FSpell: Pointer;
    FLanguage : TRVUnicodeString;
    FEncoding: String;
    procedure SetLanguage(const Value: TRVUnicodeString);
    procedure SetActive(const Value: Boolean);
    function GetActive: Boolean;
  protected
    function GetEncoding: String; virtual;
    function EncodeString(const S: TRVUnicodeString; var Res: TRVRawByteString): Boolean;
    function DecodeString(S: TRVRawByteString): TRVUnicodeString;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    function Load(Language: TRVUnicodeString = ''): Boolean;
    procedure UnLoad;
    function Spell(Word: TRVUnicodeString): Boolean;
    procedure Suggest(Word: TRVUnicodeString; Suggestions: {$IFDEF USERVTNT}TTntStrings{$ELSE}TStrings{$ENDIF});
    //procedure SuggestAuto(Word: TRVUnicodeString; Suggestions: TStrings);
    {$IFDEF RICHVIEWDEF2005}
    procedure SuggestW(Word: TRVUnicodeString; Suggestions: TWideStrings);
    //procedure SuggestAutoW(Word: TRVUnicodeString; Suggestions: TWideStrings);
    {$ENDIF}
    property Active: Boolean read GetActive write SetActive;
    property Language: TRVUnicodeString read FLanguage write SetLanguage;
    property Encoding: String read FEncoding;
  end;

  TDictionaries = class(TCollection)
  private
    FSpeller: TRVHunSpell;
    function GetItem(Index: Integer): TDictionary;
    procedure SetItem(Index: Integer; Value: TDictionary);
  protected
    function GetOwner: TPersistent; override;
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(Speller: TRVHunSpell); reintroduce; overload;
    function Add: TDictionary;
    property Items[Index: Integer]: TDictionary read GetItem write SetItem; default;
  end;

  TRVHunSpellFormEvent = procedure (Sender: TRVHunSpell; Form: TRVHunSpellCustomSpell) of object;

  TRVHunSpellFormAction = (rvhsIgnore, rvhsIgnoreAll, rvhsChange, rvhsChangeAll,
    rvhsAdd);

  TRVHunSpellFormStyle = (rvhfClassic, rvhfMSWord);

  TRVHunSpellFormActionEvent = procedure(Sender: TRVHunSpell;
    const AWord, AReplaceTo: TRVUnicodeString; Action: TRVHunSpellFormAction) of object;

  {$IFDEF RICHVIEWDEFXE2}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
  {$ENDIF}

  TRVHunSpell = class(TComponent)
  private
    FItems: TDictionaries;
    FDictDir, FDllName, FDllDir: string;
    FOnSpellFormAction: TRVHunSpellFormActionEvent;
    FOnSpellForm: TRVHunSpellFormEvent;
    FReplaceFrom, FReplaceTo, FIgnore, FCustomDic: TRVHunSpellStringList;
    FSpellFormStyle: TRVHunSpellFormStyle;
    FCustomDictionaryFileName: String;
    FDialogShown: Boolean;
    procedure SetItems(Value: TDictionaries);
    procedure SetDictDir(const Value: string);
    procedure SetDllDir(const Value: string);
    function GetActive: Boolean;
    procedure SetActive(const Value: Boolean);
    procedure DoSpellFormAction(mr: TModalResult;
      const AWord, AReplaceTo: TRVUnicodeString);
    function GetCustomDicFullFileName: String;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Spell(Word: TRVUnicodeString; Language: TRVUnicodeString = ''): Boolean;
    procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
    procedure Suggest(Word: TRVUnicodeString;
      Suggestions: {$IFDEF USERVTNT}TTntStrings{$ELSE}TStrings{$ENDIF};
      Language: TRVUnicodeString = '');
    {$IFDEF RICHVIEWDEF2005}
    procedure SuggestW(Word: TRVUnicodeString;
      Suggestions: TWideStrings; Language: TRVUnicodeString = '');
    {$ENDIF}
    procedure LoadAllDictionaries;
    procedure AddToIgnoreList(const Word: TRVUnicodeString);
    procedure AddToDictionary(const Word: TRVUnicodeString);
    procedure ResetIgnoreAndReplace;
    function LoadCustomDic: Boolean;
    procedure SaveCustomDic;
    property DialogShown: Boolean read FDialogShown;
  published
    property SpellFormStyle: TRVHunSpellFormStyle
      read FSpellFormStyle write FSpellFormStyle default rvhfClassic;
    property Active: Boolean read GetActive write SetActive;
    property Dictionaries: TDictionaries  read FItems write SetItems;
    property DictDir: string read FDictDir write SetDictDir;
    property DllDir: string read FDllDir write SetDllDir;
    property DllName: string read FDllName write FDllName;
    property OnSpellForm: TRVHunSpellFormEvent read FOnSpellForm write FOnSpellForm;
    property OnSpellFormAction: TRVHunSpellFormActionEvent read FOnSpellFormAction
        write FOnSpellFormAction;
    property CustomDictionaryFileName: String read FCustomDictionaryFileName
      write FCustomDictionaryFileName;
  end;

  TRVHunSpellHelper = class (TRVWordEnumerator)
    private
      FRVHunSpell: TRVHunSpell;
      frm : TRVHunSpellCustomSpell;
    protected
      function ProcessWord: Boolean; override;
    public
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      destructor Destroy; override;
    published
      property RVHunSpell: TRVHunSpell read FRVHunSpell write FRVHunSpell;
  end;


  procedure Register;

implementation

uses RVKoi8,
    {$IFDEF USERVTNT}
    RVTntSpellFormUtils,
    {$ELSE}
    RVSpellFormUtils,
    {$ENDIF}
    RVISO8859;

function IsPathDelimiter(const S: string; Index: Integer): Boolean;
begin
  Result := (Index > 0) and (Index <= Length(S)) and (S[Index] = '\')
    and (ByteType(S, Index) = mbSingleByte);
end;

function IncludeTrailingPathDelimiter(const S: string): string;
begin
  Result := S;
  if (S<>'') and not IsPathDelimiter(Result, Length(Result)) then
    Result := Result + '\';                           
end;


type
  PPRVAnsiChar = ^PRVAnsiChar;

var hunspell_initialize: function(aff_file: PRVAnsiChar; dict_file: PRVAnsiChar): Pointer; cdecl;
var hunspell_uninitialize: procedure(sspel: Pointer); cdecl;
var hunspell_spell: function(spell: Pointer; word: PRVAnsiChar): Boolean; cdecl;
var hunspell_suggest: function(spell: Pointer; word: PRVAnsiChar; var suggestions: PPRVAnsiChar): Integer; cdecl;
//var hunspell_suggest_auto: function(spell: Pointer; word: PRVAnsiChar; var suggestions: PPRVAnsiChar): Integer; cdecl;
var hunspell_suggest_free: procedure(spell: Pointer; suggestions: PPRVAnsiChar; suggestLen: Integer); cdecl;
var hunspell_get_dic_encoding: function(spell: Pointer): PRVAnsiChar; cdecl;
//var hunspell_put_word: function(spell: Pointer; word: PRVAnsiChar): Integer; cdecl;

var  DLLHandle: THandle;

function LoadHunspell(DllDir: string;dllname:string):boolean;
var
  ModuleName: array[0..MAX_PATH] of Char;
  OldDir: string;

function LinkFunctions:boolean;
  begin
    Result := True;
    try
      @hunspell_initialize := GetProcAddress(DLLHandle, 'hunspell_initialize');
      if not Assigned(@hunspell_initialize) then Result := False;
      @hunspell_uninitialize := GetProcAddress(DLLHandle, 'hunspell_uninitialize');
      if not Assigned(@hunspell_uninitialize) then Result := False;
      @hunspell_spell := GetProcAddress(DLLHandle, 'hunspell_spell');
      if not Assigned(@hunspell_spell) then Result := False;
      @hunspell_suggest := GetProcAddress(DLLHandle, 'hunspell_suggest');
      if not Assigned(@hunspell_suggest) then Result := False;
      //@hunspell_suggest_auto := GetProcAddress(DLLHandle, 'hunspell_suggest_auto');
      //if not Assigned(@hunspell_suggest_auto) then Result := False;
      @hunspell_suggest_free := GetProcAddress(DLLHandle, 'hunspell_suggest_free');
      if not Assigned(@hunspell_suggest_free) then Result := False;
      @hunspell_get_dic_encoding := GetProcAddress(DLLHandle, 'hunspell_get_dic_encoding');
      if not Assigned(@hunspell_get_dic_encoding) then Result := False;
      //@hunspell_put_word := GetProcAddress(DLLHandle, 'hunspell_put_word');
      //if not Assigned(@hunspell_put_word) then Result := False;
    except
      Result := False;
      exit;
    end;
  end;
begin
  Result := True;
  if DLLHandle <> 0 then Exit;
  OldDir := GetCurrentDir;
  if DllDir = '' then
  begin
    GetModuleFileName(HInstance, ModuleName, SizeOf(ModuleName));
    ChDir(ExtractFilePath(ModuleName));
  end else ChDir(DllDir);
  try
    Result := False;
    if dllname = '' then dllname := HUNSPELLDLL;

    DLLHandle := LoadLibrary(PChar(dllname));
    if DLLHandle <> 0 then Result := LinkFunctions;
  except
    Result := False;
  end;
  ChDir(OldDir);
  if not Result then
    DLLHandle := 0;
end;

procedure Register;
begin
  RegisterComponents('RichView', [TRVHunSpell]);
end;

{==================================== TDictionary =============================}

constructor TDictionary.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FSpell := nil;
  FLanguage := '';
  FEncoding := '';
end;
{------------------------------------------------------------------------------}
{ The following encodings can be found in available Hunspell dictionaries:
 ISO8859-1
 ISO8859-2
 ISO8859-3
 ISO8859-4
 ISO8859-6
 ISO8859-7
 ISO8859-8
 ISO8859-9
 ISO8859-13
 ISO8859-14
 iso8859-15
 ISO8859-15
 KOI8-R
 KOI8-U
 microsoft-cp1251
 TIS620-2533 (!!!)
 UTF-8
}
{------------------------------------------------------------------------------}
// Decoding from dictionary format to Unicode
function TDictionary.DecodeString(S: TRVRawByteString): TRVUnicodeString;
{$IFDEF RVUNICODESTR}
var AnsiS: RawByteString;
{$ELSE}
var Len: Integer;
{$ENDIF}
var v: Integer;
begin
  if FEncoding = 'UTF-8' then
  begin
    {$IFDEF RICHVIEWDEF2009}
    Result := UTF8ToString(S);
    {$ELSE}
    Result := UTF8Decode(S);
    {$ENDIF}
    Exit;
  end;
  if (FEncoding = 'KOI8-R') or (FEncoding = 'KOI8-U') then
  begin
    ConvertKoi8ToWin(S);
    {$IFDEF RVUNICODESTR}
    AnsiS := S;
    SetCodePage(AnsiS, 1251, False);
    Result := String(AnsiS);
    {$ELSE}
    Len := MultiByteToWideChar(1251, 0, Pointer(S), Length(S), nil, 0);
    SetLength(Result, Len);
    MultiByteToWideChar(1251, 0, Pointer(S), Length(S),
      Pointer(Result), Length(Result));
    {$ENDIF}
    Exit;
  end;
  if Copy(Encoding, 1, 8)='ISO8859-' then begin
    v := StrToIntDef(Copy(Encoding, 9, 2), 0);
    ConvertISO8859AndUnicode(S, Result, v, True);
    exit;
  end;
  if Copy(Encoding, 1, 12)='microsoft-cp' then begin
    v := StrToIntDef(Copy(Encoding, 13, 4), CP_ACP);
    {$IFDEF RVUNICODESTR}
    AnsiS := S;
    SetCodePage(AnsiS, v, False);
    Result := String(AnsiS);
    {$ELSE}
    Len := MultiByteToWideChar(v, 0, Pointer(S), Length(S), nil, 0);
    SetLength(Result, Len);
    MultiByteToWideChar(v, 0, Pointer(S), Length(S),
      Pointer(Result), Length(Result));
    {$ENDIF}
    exit;
  end;
  Result := String(TRVAnsiString(S));
end;
{------------------------------------------------------------------------------}
// Encoding from Unicode to dictionary format
function TDictionary.EncodeString(const S: TRVUnicodeString;
  var Res: TRVRawByteString): Boolean;
var Len, v: Integer;
    DefChar: TRVAnsiChar;
    UsedDefChar: LongBool;
    S2: TRVUnicodeString;
begin
  if FEncoding = 'UTF-8' then
  begin
    Res := UTF8Encode(S);
    Result := True;
    Exit;
  end;
  if (FEncoding = 'KOI8-R') or (FEncoding = 'KOI8-U') then begin
    Len := WideCharToMultiByte(1251, 0, Pointer(S), Length(S), nil, 0, nil, nil);
    SetLength(Res, Len);
    DefChar := '?';
    UsedDefChar := False;
    WideCharToMultiByte(1251, 0, Pointer(S), Length(S),
      PAnsiChar(Res), Length(Res), @DefChar, @UsedDefChar);
    Result := not UsedDefChar;
    ConvertWinToKoi8(Res);
    Exit;
  end;
  if Copy(Encoding, 1, 8)='ISO8859-' then begin
    v := StrToIntDef(Copy(Encoding, 9, 2), 0);
    S2 := S;
    Result := ConvertISO8859AndUnicode(Res, S2, v, False);
    exit;
  end;
  if Copy(Encoding, 1, 12)='microsoft-cp' then begin
    v := StrToIntDef(Copy(Encoding, 13, 4), CP_ACP);
    Len := WideCharToMultiByte(v, 0, Pointer(S), Length(S), nil, 0, nil, nil);
    SetLength(Res, Len);
    DefChar := '?';
    UsedDefChar := False;
    WideCharToMultiByte(v, 0, Pointer(S), Length(S),
      PAnsiChar(Res), Length(Res), @DefChar, @UsedDefChar);
    Result := not UsedDefChar;
    exit;
  end;
  Res := TRVAnsiString(S);
  Result := True;
end;
{------------------------------------------------------------------------------}
destructor TDictionary.Destroy;
begin
  UnLoad;
  inherited;
end;
{------------------------------------------------------------------------------}
function TDictionary.GetActive: Boolean;
begin
  Result := Assigned(FSpell);
end;
{------------------------------------------------------------------------------}
function TDictionary.GetEncoding: String;
begin
  Result := FEncoding;
end;
{------------------------------------------------------------------------------}
function TDictionary.Load(Language: TRVUnicodeString): Boolean;
var
  dic_file: TRVAnsiString;
  aff_file: TRVAnsiString;
  ModuleName: array[0..MAX_PATH] of Char;
  OldDir: string;
  IsoVersion: Integer;
  HunSpell: TRVHunSpell;
begin
  Result := True;
  // if the Language parameter is not specified, using a preset value
  if Language = '' then Language := FLanguage;
  // If the dictionary already loaded
  if FSpell <> nil then
  begin
    // If the requested dictionary already loaded, exiting
    if FLanguage = Language then Exit
    // we need to unload the old dictionary
    else Unload;
  end;
  // Dictionary language does not depend on loaded/not loaded state
  FLanguage := Language;
  // If language is undefined, exiting
  if FLanguage = '' then Exit;
  Result := False;
  // Initializing local variables
  FEncoding := '';
  FSpell := nil;
  OldDir := GetCurrentDir;
  HunSpell := (Collection as TDictionaries).FSpeller;
  // If DLL cannot be loaded, exiting
  if not LoadHunspell(IncludeTrailingPathDelimiter(HunSpell.DllDir), HunSpell.DllName) then Exit;
  try
    // Moving to the directory of the module (because relative paths are defined relative to it)
    GetModuleFileName(HInstance, ModuleName, SizeOf(ModuleName));
    ChDir(ExtractFilePath(ModuleName));
    // Making dictionary file names
    dic_file := TRVAnsiString(IncludeTrailingPathDelimiter(HunSpell.DictDir) + string(Language) + '.dic');
    aff_file := TRVAnsiString(IncludeTrailingPathDelimiter(HunSpell.FDictDir) + string(Language) + '.aff');
    FSpell := hunspell_initialize(PRVAnsiChar(aff_file),PRVAnsiChar(dic_file));
    // If the dictionary cannot be loaded, exiting
    if not Assigned(FSpell) then Exit;
    // Getting the library encoding
    FEncoding := String(hunspell_get_dic_encoding(FSpell));
    if Copy(Encoding, 1, 8)='ISO8859-' then begin
      IsoVersion := StrToIntDef(Copy(Encoding, 9, 2), 0);
      AllocateISO8859Table(IsoVersion);
    end;
    Result := True;
  except
    FSpell := nil;
    Result := False;
  end;
  ChDir(OldDir);
  // If the loading succeeded, assigning dictionary language
  if Result then FLanguage := Language
  else FLanguage := '';
end;
{------------------------------------------------------------------------------}
procedure TDictionary.SetActive(const Value: Boolean);
begin
  if Value = Active then Exit;
  if Value then Load else UnLoad;
end;
{------------------------------------------------------------------------------}
procedure TDictionary.SetLanguage(const Value: TRVUnicodeString);
begin
  if FLanguage <> Value then FLanguage := Value;
  if not Assigned(FSpell) then Load(Value);
end;
{------------------------------------------------------------------------------}
function TDictionary.Spell(Word: TRVUnicodeString): Boolean;
var s: TRVRawByteString;
begin
  Result := False;
  try
    if not Assigned(FSpell) then Exit;
    Result := EncodeString(Word, s);
    if Result then
      Result := hunspell_spell(FSpell, PRVAnsiChar(s));
  except
    Result := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TDictionary.Suggest(Word: TRVUnicodeString; Suggestions: {$IFDEF USERVTNT}TTntStrings{$ELSE}TStrings{$ENDIF});
var
  I: Integer;
  Len: Integer;
  wrds: PPRVAnsiChar;
  s: TRVRawByteString;
  Converted: Boolean;
  AWord: TRVUnicodeString;
begin
  if not Assigned(FSpell) then Exit;
  // Suggestions are not cleared, to allow using several dictionaries
  Suggestions.BeginUpdate;
  try
    Converted := EncodeString(Word, s);
    if Converted then begin
      Len := hunspell_suggest(FSpell, PRVAnsiChar(s), wrds);
      for I := 1 to Len do begin
        AWord := DecodeString(wrds^);
        if Suggestions.IndexOf(AWord)<0 then
          Suggestions.Add(AWord);
        {$IFDEF RICHVIEWDEFXE2}
        Inc(NativeInt(wrds), SizeOf(Pointer));
        {$ELSE}
        Inc(Integer(wrds), SizeOf(Pointer));        
        {$ENDIF}
      end;
      {$IFDEF RICHVIEWDEFXE2}
      Dec(NativeInt(wrds), SizeOf(Pointer) * Len);
      {$ELSE}
      Dec(Integer(wrds), SizeOf(Pointer) * Len);      
      {$ENDIF}
      hunspell_suggest_free(FSpell, wrds, Len);
    end;
  finally
    Suggestions.EndUpdate;
  end;
end;
{------------------------------------------------------------------------------}
{
procedure TDictionary.SuggestAuto(Word: TRVUnicodeString; Suggestions: TStrings);
var
  I: Integer;
  Len: Integer;
  wrds: PPRVAnsiChar;
  s: TRVRawByteString;
  Converted: Boolean;
begin
  if not Assigned(FSpell) then Exit;
  Suggestions.BeginUpdate;
  try
    Converted := EncodeString(Word, s);
    if Converted then begin
      Len := hunspell_suggest_auto(FSpell, PRVAnsiChar(s), wrds);
      for I := 1 to Len do begin
        Suggestions.Add(DecodeString(wrds^));
        Inc(Integer(wrds), SizeOf(Pointer));
      end;
      Dec(Integer(wrds), SizeOf(Pointer) * Len);
      hunspell_suggest_free(FSpell, wrds, Len);
    end;
  finally
    Suggestions.EndUpdate;
  end;
end;
}
{$IFDEF RICHVIEWDEF2005}
{------------------------------------------------------------------------------}
procedure TDictionary.SuggestW(Word: TRVUnicodeString; Suggestions: TWideStrings);
var
  I: Integer;
  Len: Integer;
  wrds: PPRVAnsiChar;
  s: TRVRawByteString;
  Converted: Boolean;
  AWord: TRVUnicodeString;
begin
  if not Assigned(FSpell) then Exit;
  // Suggestions are not cleared, to allow using several dictionaries
  Suggestions.BeginUpdate;
  try
    Converted := EncodeString(Word, s);
    if Converted then begin
      Len := hunspell_suggest(FSpell, PRVAnsiChar(s), wrds);
      for I := 1 to Len do begin
        AWord := DecodeString(wrds^);
        if Suggestions.IndexOf(AWord)<0 then
          Suggestions.Add(AWord);
        Inc(NativeInt(wrds), SizeOf(Pointer));
      end;
      Dec(NativeInt(wrds), SizeOf(Pointer) * Len);
      hunspell_suggest_free(FSpell, wrds, Len);
    end;
  finally
    Suggestions.EndUpdate;
  end;
end;
{------------------------------------------------------------------------------}
{
procedure TDictionary.SuggestAutoW(Word: TRVUnicodeString; Suggestions: TWideStrings);
var
  I: Integer;
  Len: Integer;
  wrds: PPRVAnsiChar;
  s: TRVRawByteString;
  Converted: Boolean;
begin
  if not Assigned(FSpell) then Exit;
  Suggestions.BeginUpdate;
  try
    Converted := EncodeString(Word, s);
    if Converted then begin
      Len := hunspell_suggest_auto(FSpell, PRVAnsiChar(s), wrds);
      for I := 1 to Len do begin
        Suggestions.Add(DecodeString(wrds^));
        Inc(Integer(wrds), SizeOf(Pointer));
      end;
      Dec(Integer(wrds), SizeOf(Pointer) * Len);
      hunspell_suggest_free(FSpell, wrds, Len);
    end;
  finally
    Suggestions.EndUpdate;
  end;
end;
}
{$ENDIF}
{------------------------------------------------------------------------------}
procedure TDictionary.UnLoad;
begin
  if FSpell <> nil then hunspell_uninitialize(FSpell);
  FSpell := nil;
end;
{=============================== TDictionaries ================================}
function TDictionaries.Add: TDictionary;
begin
  Result := TDictionary(inherited Add);
end;
{------------------------------------------------------------------------------}
constructor TDictionaries.Create(Speller: TRVHunSpell);
begin
  inherited Create(TDictionary);
  FSpeller := Speller;
end;
{------------------------------------------------------------------------------}
function TDictionaries.GetItem(Index: Integer): TDictionary;
begin
  Result := TDictionary(inherited GetItem(Index));
end;
{------------------------------------------------------------------------------}
function TDictionaries.GetOwner: TPersistent;
begin
  Result := FSpeller;
end;
{------------------------------------------------------------------------------}
procedure TDictionaries.AssignTo(Dest: TPersistent);
var i: Integer;
begin
  if Dest is TStrings then begin
    TStrings(Dest).BeginUpdate;
    TStrings(Dest).Clear;
    try
      for i := 0 to Count - 1 do
        TStrings(Dest).Add(Items[i].Language);
    finally
      TStrings(Dest).EndUpdate;
    end;
    end
  else
    inherited;
end;
{------------------------------------------------------------------------------}
procedure TDictionaries.SetItem(Index: Integer; Value: TDictionary);
begin
  inherited SetItem(Index, Value);
end;

{==================================== TRVHunSpell =============================}

constructor TRVHunSpell.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TDictionaries.Create(Self);
  FReplaceFrom := TRVHunSpellStringList.Create;
  FReplaceTo   := TRVHunSpellStringList.Create;
  FIgnore      := TRVHunSpellStringList.Create;
  FCustomDic   := TRVHunSpellStringList.Create;
  CustomDictionaryFileName := DEFCUSTOMDICNAME;
  DLLName := HUNSPELLDLL;
  {$IFDEF RICHVIEWDEF6}
  FIgnore.Sorted := True;
  FIgnore.CaseSensitive := True;
  FCustomDic.Sorted := True;
  FCustomDic.CaseSensitive := True;
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
destructor TRVHunSpell.Destroy;
begin
  SaveCustomDic;
  FReplaceFrom.Free;
  FReplaceTo.Free;
  FIgnore.Free;
  FItems.Free;
  FCustomDic.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.DoSpellFormAction(mr: TModalResult; const AWord,
  AReplaceTo: TRVUnicodeString);
var Action: TRVHunSpellFormAction;
begin
  if not Assigned(FOnSpellFormAction) then
    exit;
  case mr of
    mrNo:
      Action := rvhsIgnore;
    mrNoToAll:
      Action := rvhsIgnoreAll;
    mrOk:
      Action := rvhsAdd;
    mrYes:
      Action := rvhsChange;
    mrYesToAll:
      Action := rvhsChangeAll;
    else
      exit;
  end;
  FOnSpellFormAction(Self, AWord, AReplaceTo, Action);
end;
{------------------------------------------------------------------------------}
function TRVHunSpell.GetActive: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Dictionaries.Count - 1 do
    if not Dictionaries[I].Active then Exit;
  Result := True;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.SetActive(const Value: Boolean);
var
  I: Integer;
begin
  //  This property shows the state of all dictionaries, and it is True only
  // if all collection items are active,
  // so, if it is False, some items may be still active,
  // so we do not compare the parameter with the existing values

  for I := 0 to Dictionaries.Count - 1 do
    Dictionaries[I].Active := Value;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.SetDictDir(const Value: string);
begin
  if FDictDir = Value then Exit;
  FDictDir := Value;
  // The dictionaries directory is changed, so we unloading
  // all dictionaties
  Active := False;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.SetDllDir(const Value: string);
begin
  if FDllDir = Value then Exit;
  FDllDir := Value;
end;

{------------------------------------------------------------------------------}
procedure TRVHunSpell.SetItems(Value: TDictionaries);
begin
  FItems.Assign(Value);
end;
{------------------------------------------------------------------------------}
function TRVHunSpell.Spell(Word: TRVUnicodeString; Language: TRVUnicodeString): Boolean;
var
  I: Integer;
begin
  {$IFDEF RICHVIEWDEF6}
  Result := FIgnore.Find(Word, I) or FCustomDic.Find(Word, I);
  {$ELSE}
  Result := (FIgnore.IndexOf(Word)>=0) or (FCustomDic.IndexOf(Word)>=0);
  {$ENDIF}
  if Result then
    exit;
  for I := 0 to Dictionaries.Count - 1 do
  begin
    if (Language = '') or (Language = Dictionaries[I].FLanguage) then
      Result := Dictionaries[I].Spell(Word);
    if Result then Break;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
var Helper: TRVHunSpellHelper;
begin
  Helper := TRVHunSpellHelper.Create;
  try
    Helper.RVHunSpell := Self;
    Helper.Unicode := True;
    Helper.Check(Editor, Scope);
    FDialogShown := Helper.DialogShown;
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.Suggest(Word: TRVUnicodeString;
  Suggestions: {$IFDEF USERVTNT}TTntStrings{$ELSE}TStrings{$ENDIF};
  Language: TRVUnicodeString);
var
  I: Integer;
begin
  Suggestions.Clear;
  for I := 0 to Dictionaries.Count - 1 do
  begin
    if (Language = '') or (Language = Dictionaries[I].FLanguage) then
      Dictionaries[I].Suggest(Word, Suggestions);
  end;
end;
{------------------------------------------------------------------------------}
{
procedure TRVHunSpell.SuggestAuto(Word: TRVUnicodeString; Suggestions: TStrings;
  Language: TRVUnicodeString);
var
  I: Integer;
begin
  Suggestions.Clear;
  for I := 0 to Dictionaries.Count - 1 do
  begin
    if (Language = '') or (Language = Dictionaries[I].FLanguage) then
      Dictionaries[I].SuggestAuto(Word, Suggestions);
  end;
end;
}
{------------------------------------------------------------------------------}
{$IFDEF RICHVIEWDEF2005}
procedure TRVHunSpell.SuggestW(Word: TRVUnicodeString; Suggestions: TWideStrings;
  Language: TRVUnicodeString);
var
  I: Integer;
begin
  Suggestions.Clear;
  for I := 0 to Dictionaries.Count - 1 do
  begin
    if (Language = '') or (Language = Dictionaries[I].FLanguage) then
      Dictionaries[I].SuggestW(Word, Suggestions);
  end;
end;
{------------------------------------------------------------------------------}
{
procedure TRVHunSpell.SuggestAutoW(Word: TRVUnicodeString; Suggestions: TWideStrings;
  Language: TRVUnicodeString);
var
  I: Integer;
begin
  Suggestions.Clear;
  for I := 0 to Dictionaries.Count - 1 do
  begin
    if (Language = '') or (Language = Dictionaries[I].FLanguage) then
      Dictionaries[I].SuggestAutoW(Word, Suggestions);
  end;
end;
}
{$ENDIF}
{------------------------------------------------------------------------------}
procedure TRVHunSpell.LoadAllDictionaries;
var sr: TSearchRec;
begin
  if FindFirst(IncludeTrailingPathDelimiter(DictDir)+'*.dic', 0, sr)=0 then begin
    repeat
      with Dictionaries.Add do begin
        Load(Copy(sr.Name, 1, Length(sr.Name)-4));
        if not Active then
          Free;
      end;
    until FindNext(sr)<>0;
    FindClose(sr);
  end;
  LoadCustomDic;  
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.AddToIgnoreList(const Word: TRVUnicodeString);
begin
  FIgnore.Add(Word);
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.AddToDictionary(const Word: TRVUnicodeString);
begin
  FCustomDic.Add(Word);
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.ResetIgnoreAndReplace;
begin
  FIgnore.Clear;
  FReplaceFrom.Clear;
  FReplaceTo.Clear;
end;
{------------------------------------------------------------------------------}
function TRVHunSpell.GetCustomDicFullFileName: String;
begin
  Result := FCustomDictionaryFileName;
  if Result='' then
    Result := DEFCUSTOMDICNAME;
  if (Pos(':', Result)=0) and (Result[1]<>'\') and (Result[1]<>'/') then
    Result := ExtractFilePath(Application.ExeName)+Result;
end;
{------------------------------------------------------------------------------}
function TRVHunSpell.LoadCustomDic: Boolean;
var FileName: String;
  Stream: TFileStream;
  s: TRVUnicodeString;
begin
  Result := True;
  FileName := GetCustomDicFullFileName;
  if not FileExists(FileName) then
    exit;
  try
    Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      FCustomDic.Clear;
      Result := Stream.Size mod 2 = 0;
      if Result then begin
        SetLength(s, Stream.Size div 2);
        Stream.ReadBuffer(Pointer(s)^, Stream.Size);
      end;
    finally
      Stream.Free;
    end;
    FCustomDic.Text := s;
  except
    Result := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVHunSpell.SaveCustomDic;
var FileName: String;
  Stream: TFileStream;
  s: TRVUnicodeString;
begin
  FileName := GetCustomDicFullFileName;
  if FCustomDic.Count=0 then begin
    DeleteFile(FileName);
    exit;
  end;
  try
    Stream := TFileStream.Create(FileName, fmCreate);
    try
      s := FCustomDic.Text;
      Stream.WriteBuffer(Pointer(s)^, Length(s)*2);
    finally
      Stream.Free;
    end;
  except
    ;
  end;
end;
{================================ TRVHunSpellHelper ===========================}
procedure TRVHunSpellHelper.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
begin
  Run(Editor, Scope);
end;
{------------------------------------------------------------------------------}
destructor TRVHunSpellHelper.Destroy;
begin
  frm.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
function TRVHunSpellHelper.ProcessWord: Boolean;
var Index: Integer;
    mr: TModalResult;
    IsFormCreated: Boolean;
    SuggestionsStrings: {$IFDEF USERVTNT}TTntStrings{$ELSE}TStrings{$ENDIF};
    Correction: WideString;
begin
  Result := True;
  {$IFDEF RICHVIEWDEF6}
  if RVHunSpell.FIgnore.Find(FWordW, Index) or
    RVHunSpell.FCustomDic.Find(FWordW, Index) then
  {$ELSE}
  if (RVHunSpell.FIgnore.IndexOf(FWordW)>=0) or
    (RVHunSpell.FCustomDic.IndexOf(FWordW)>=0) then
  {$ENDIF}
    exit;

  Index := RVHunSpell.FReplaceFrom.IndexOf(FWordW);
  if Index>=0 then begin
    FWordChanged := True;
    FWordW := RVHunSpell.FReplaceTo[Index];
    exit;
  end;
  if FRVHunSpell.Spell(FWordW) then
    exit;
  SelectCurrentWord;
  IsFormCreated := frm=nil;
  CreateSpellForm(TRVSpellFormStyle(RVHunSpell.SpellFormStyle), frm);
  if IsFormCreated and Assigned(RVHunSpell.OnSpellForm) then
    RVHunSpell.OnSpellForm(RVHunSpell, frm);
  SuggestionsStrings := frm.GetStringsForSuggestions;
  SuggestionsStrings.BeginUpdate;
  try
    SuggestionsStrings.Clear;
    RVHunSpell.Suggest(FWordW, SuggestionsStrings);
  finally
    SuggestionsStrings.EndUpdate;
  end;
  BeforeShowSpellForm(FWordW, FTextW, FStartOffs, FEndOffs, frm,
    FRVData.GetRVStyle.TextStyles[FRVData.GetItemStyle(FItemNo)]);
  FDialogShown := True;
  mr := frm.ShowModal;
  Correction := frm.GetCorrectedText;
  FRVHunSpell.DoSpellFormAction(mr, FWordW, Correction);
  case mr of
    mrNo: ;
    mrNoToAll:
      RVHunSpell.FIgnore.Add(FWordW);
    mrOk:
      RVHunSpell.FCustomDic.Add(FWordW);
    mrYes, mrYesToAll:
      begin
        if mr=mrYesToAll then begin
          RVHunSpell.FReplaceFrom.Add(FWordW);
          RVHunSpell.FReplaceTo.Add(Correction);
        end;
        if not frm.IsReplacingItemText then begin
          FWordW := Correction;
          FWordChanged := True;
          end
        else begin
          FTextW := Correction;
          FTextChanged := True;
        end;
      end;
    mrCancel:
      Result := False;
  end;
end;



initialization
  DLLHandle := 0;
finalization
  if DLLHandle<>0 then
    FreeLibrary(DLLHandle);

end.