
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVLSSpell: component for                       }
{       spell-checking RichViewEdit using               }
{       LS Speller                                      }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       LS Spell : Copyright (c)  Luzius Schneider      }
{         based on former work of Alexander Obukhov     }
{       luzius.schneider@switzerland.org                }
{       http://www.luziusschneider.com/SpellerHome.htm  }
{                                                       }
{*******************************************************}

{
  About LS Speller:

  TSpellChecker is a non-visual Delphi component designed to add spell check
  capability to any application.
  It uses ISpell dictionaries installed with "ISpell - LS-Distribution".
  Additionally, TSpellChecker can use spell check dictionaries installed with
  MSOffice 95 or 97. For higher versions of MSOffice, only some languages
  will work.

  Installing this component:

    - Add this file (RVLSSpell.pas) in some package. Compile this package.
      TRVLSSpell component will appear on "RichView" page of the
      Component Palette.

  Using:
    - place TRVLSSpell on the form.
    - assign TSpellChecker to RVLSSpell.SpellChecker
    - in program, call TRVLSSpell.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call TRVLSSpell.Check(rve, rvesFromStart) to check from the
      beginning.
  Methods:
    - ClearIgnoreAll: clears a list of ignored words (this list is not
      cleared between calls of Check)
    - ClearChangeAll: clears a list of autoreplaced words (this list is not
      cleared between calls of Check)
  Properties:
    - SpellFormStyle defines which dialog style is used
      (classic or like in MS Word)  
  Events:
    - OnSpellForm - event occuring when spell form is created. Can be used for
      localization.

  About ISpell's Russian and Ukrainian dictionaries:
    These dictionaries have Koi8 encoding. This component performs ANSI<->Koi8
    conversion automatically.
    Because LS' TSpellChecker bug (it's unaware that Koi8 should be used),
    spoIgnoreAllCaps must be excluded from TSpellChecker.Options if you want to
    use these dictionaries.
    Note: MS Word dictionaries have ANSI encoding.

  Functions for live spelling:
    - checking word in RichViewEdit.OnSpellingCheck: TRVLSSpell.IsMisspelled
    - adding to ignore list: TRVLSSpell.AddToIgnoreList
    - adding to dictionary: TRVLSSpell.SpellChecker.AddWord
      (TRVLSSpell.GetCurLanguage can be used as Language parameter)
    - suggestions: TRVLSSpell.SuggestWords
    - event on adding word to ignore list: TRVLSSpell.OnSpellFormAction
    - event on adding word to dictionary: TRVLSSpell.OnSpellFormAction

  History:
  2011-03-20: property DialogShown specifies if a spelling check dialog has been
    shown during the last spelling check.
  2004-04-18:
    - removed: Language and DefaultLanguage properties.
    - change: the component tries to use language of text (determined by Charset);
      if charset is default, it uses TRVLSSpell.SpellChecker.Language
      (type of dictionary is always defined by TRVLSSpell.SpellChecker.SpellerType)
  2004-04-07: live spelling and Koi8


  Known problems:
  - LS speller may raise exception in destructor. It's handled and looks harmless.
}

unit RVLSSpell;

interface
{$I RV_Defs.inc}

uses
  SysUtils, Classes, Controls,Forms,
  Spellers, Langs, RVKoi8,
  RVStyle, RVEdit, RVUni, RVTypes,
  RVWordEnum, RVCustomSpellFrm;

type
  TRVLSSpell = class;

  TRVLSSpellFormStyle = (rvlsfClassic, rvlsfMSWord);

  TOnSpellFormEvent = procedure (Sender: TRVLSSpell; Form: TfrmRVCustomSpell) of object;
  TRVLSSpellFormAction = (rvlssIgnore, rvlssIgnoreAll, rvlssChange, rvlssChangeAll,
    rvlssAdd);
  TRVLSSpellFormActionEvent = procedure(Sender: TRVLSSpell;
    const AWord, AReplaceTo: String; Action: TRVLSSpellFormAction) of object;

  TIgnoreCollectionItem = class (TCollectionItem)
    public
      Word: String;
      Language: TLanguage;
  end;
  TReplaceCollectionItem = class (TCollectionItem)
    public
      WordFrom, WordTo: String;
      Language: TLanguage;
  end;

  TIgnoreCollection = class (TCollection)
    constructor Create;
    function Find(const Word: String; Language: TLanguage): Boolean;
    procedure Add(const Word: String; Language: TLanguage);
  end;

  TReplaceCollection = class (TCollection)
    constructor Create;
    function Find(const WordFrom: String; Language: TLanguage; var WordTo: String): Boolean;
    procedure Add(const WordFrom, WordTo: String; Language: TLanguage);
  end;

  TRVLSSpell = class (TComponent)
    private
      FOnSpellForm: TOnSpellFormEvent;
      FSpellChecker: TSpellChecker;
      FReplaceAll: TReplaceCollection;
      FIgnoreAll: TIgnoreCollection;
      FOnSpellFormAction: TRVLSSpellFormActionEvent;
      FSpellFormStyle: TRVLSSpellFormStyle;
      FDialogShown: Boolean;
      function IsCyrillic(Language: TLanguage): Boolean;
      procedure DoSpellFormAction(mr: TModalResult; const AWord,
        AReplaceTo: String);
    protected
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      procedure ClearIgnoreAll;
      procedure ClearChangeAll;
      procedure AddToIgnoreList(AWord: String; ATextStyle: TFontInfo);
      function IsMisspelled(AWord: String; ATextStyle: TFontInfo): Boolean;
      procedure SuggestWords(AWord: String; ATextStyle: TFontInfo; ASuggestions: TStrings);
      function GetCurLanguage(Style: TFontInfo): TLanguage;
      property DialogShown: Boolean read FDialogShown;
    published
      property SpellFormStyle: TRVLSSpellFormStyle
        read FSpellFormStyle write FSpellFormStyle default rvlsfClassic;
      property OnSpellForm: TOnSpellFormEvent read FOnSpellForm write FOnSpellForm;
      property SpellChecker: TSpellChecker read FSpellChecker write FSpellChecker;
      property OnSpellFormAction: TRVLSSpellFormActionEvent read FOnSpellFormAction
        write FOnSpellFormAction;
  end;

  TRVLSSpellHelper = class (TRVWordEnumerator)
    private
      FRVLSSpell: TRVLSSpell;
      frm : TfrmRVCustomSpell;
    protected
      function ProcessWord: Boolean; override;
    public
      destructor Destroy; override;
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
    published
      property RVLSSpell: TRVLSSpell read FRVLSSpell write FRVLSSpell;
  end;

  ERVLSSpellError = class (Exception);

  procedure Register;

implementation
uses RVSpellFormUtils;
// Note: It's possible to implement these collections more efficiently, using
// sorting. 
{============================== TIgnoreCollection =============================}
constructor TIgnoreCollection.Create;
begin
  inherited Create(TIgnoreCollectionItem);
end;
{------------------------------------------------------------------------------}
procedure TIgnoreCollection.Add(const Word: String; Language: TLanguage);
var Item: TIgnoreCollectionItem;
begin
  Item := TIgnoreCollectionItem(inherited Add);
  Item.Word := Word;
  Item.Language := Language;
end;
{------------------------------------------------------------------------------}
function TIgnoreCollection.Find(const Word: String;
  Language: TLanguage): Boolean;
var i: Integer;
begin
  Result := True;
  for i := 0 to Count-1 do
    if (TIgnoreCollectionItem(Items[i]).Word=Word) and
       (TIgnoreCollectionItem(Items[i]).Language=Language) then
      exit;
  Result := False;
end;
{=============================== TReplaceCollection ===========================}
constructor TReplaceCollection.Create;
begin
  inherited Create(TReplaceCollectionItem);
end;
{------------------------------------------------------------------------------}
procedure TReplaceCollection.Add(const WordFrom, WordTo: String;
  Language: TLanguage);
var Item: TReplaceCollectionItem;
begin
  Item := TReplaceCollectionItem(inherited Add);
  Item.WordFrom := WordFrom;
  Item.WordTo := WordTo;
  Item.Language := Language;
end;
{------------------------------------------------------------------------------}
function TReplaceCollection.Find(const WordFrom: String;
  Language: TLanguage; var WordTo: String): Boolean;
var i: Integer;
begin
  Result := True;
  for i := 0 to Count-1 do
    if (TReplaceCollectionItem(Items[i]).WordFrom=WordFrom) and
       (TReplaceCollectionItem(Items[i]).Language=Language) then begin
      WordTo := TReplaceCollectionItem(Items[i]).WordTo;
      exit;
    end;
  Result := False;
end;
{=================================== TRVLSSpell ================================}
constructor TRVLSSpell.Create(AOwner: TComponent);
begin
  inherited;
  FReplaceAll := TReplaceCollection.Create;
  FIgnoreAll  := TIgnoreCollection.Create;
end;
{------------------------------------------------------------------------------}
destructor TRVLSSpell.Destroy;
begin
  FReplaceAll.Free;
  FIgnoreAll.Free;
  inherited Destroy;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.ClearIgnoreAll;
begin
  FIgnoreAll.Clear;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.ClearChangeAll;
begin
  FReplaceAll.Clear;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var Helper: TRVLSSpellHelper;
begin
  Helper := TRVLSSpellHelper.Create;
  try
    Helper.RVLSSpell := Self;
    Helper.Check(Editor, Scope);
    FDialogShown := Helper.DialogShown;
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation=opRemove) and (AComponent=SpellChecker) then
    SpellChecker := nil;
end;
{------------------------------------------------------------------------------}
function TRVLSSpell.GetCurLanguage(Style: TFontInfo): TLanguage;
begin
  {$IFDEF RVLANGUAGEPROPERTY}
  Result := Style.Language;
  {$ELSE}
  Result := RVU_Charset2Language(Style.Charset);
  {$ENDIF}
  if (Result=$0000) or (Result=$0400) and (FSpellChecker<>nil) then
    Result := FSpellChecker.Language;
end;
{------------------------------------------------------------------------------}
function TRVLSSpell.IsMisspelled(AWord: String;
  ATextStyle: TFontInfo): Boolean;
var Language: TLanguage;
begin
  if Length(AWord)<=1 then begin
    Result := False;
    exit;
  end;
  Language := GetCurLanguage(ATextStyle);
  Result := not FIgnoreAll.Find(AWord, Language);
  if Result then begin
    if (SpellChecker.SpellerType=sptISpell) and IsCyrillic(Language) then
    ConvertWinToKoi8(TRVRawByteString(AWord));
    Result := not SpellChecker.IsKnownWord(AWord, Language);
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.SuggestWords(AWord: String; ATextStyle: TFontInfo;
  ASuggestions: TStrings);
var Language: TLanguage;
    i: Integer;
    Koi8: Boolean;
    s: TRVRawByteString;
begin
  ASuggestions.Clear;
  if not IsMisspelled(AWord, ATextStyle) then
    exit;
  Language := GetCurLanguage(ATextStyle);
  Koi8 := (SpellChecker.SpellerType=sptISpell) and IsCyrillic(Language);
  if Koi8 then
    ConvertWinToKoi8(TRVRawByteString(AWord));
  SpellChecker.GetVariants(AWord, ASuggestions, Language);
  if Koi8 then
    for i := 0 to ASuggestions.Count-1 do begin
      s := ASuggestions[i];
      ConvertKoi8ToWin(s);
      ASuggestions[i] := s;
    end;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.AddToIgnoreList(AWord: String; ATextStyle: TFontInfo);
var Language: TLanguage;
begin
  Language := GetCurLanguage(ATextStyle);
  if not FIgnoreAll.Find(AWord, Language) then
    FIgnoreAll.Add(AWord, Language);
end;
{------------------------------------------------------------------------------}
function TRVLSSpell.IsCyrillic(Language: TLanguage): Boolean;
begin
  Result := (Language=$0419) or (Language=$0422); // Russian or Ukrainian
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpell.DoSpellFormAction(mr: TModalResult; const AWord, AReplaceTo: String);
var Action: TRVLSSpellFormAction;
begin
  if not Assigned(FOnSpellFormAction) then
    exit;
  case mr of
    mrNo:
      Action := rvlssIgnore;
    mrNoToAll:
      Action := rvlssIgnoreAll;
    mrOk:
      Action := rvlssAdd;
    mrYes:
      Action := rvlssChange;
    mrYesToAll:
      Action := rvlssChangeAll;
    else
      exit;
  end;
  FOnSpellFormAction(Self, AWord, AReplaceTo, Action);
end;
{============================= TRVLSSpellHelper ================================}
destructor TRVLSSpellHelper.Destroy;
begin
  frm.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVLSSpellHelper.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
begin
  Run(Editor, Scope);
end;
{------------------------------------------------------------------------------}
function TRVLSSpellHelper.ProcessWord: Boolean;
var Language: TLanguage;
    mr: TModalResult;
    WordTo: String;
    Word2: String;
    i: Integer;
    IsFormCreated: Boolean;
    SuggestionsStrings: TStrings;
    Correction: String;
//    CodePage: TRVCodePage;
begin
  Result := True;
  if Length(FWordA)<=1 then
    exit;
  Language := RVLSSpell.GetCurLanguage(FFontStyle);
  if RVLSSpell.FIgnoreAll.Find(FWordA, Language) then
    exit;
  if RVLSSpell.FReplaceAll.Find(FWordA, Language, WordTo) then begin
    FWordA := WordTo;
    FWordChanged := True;
    exit;
  end;
  Word2 := FWordA;
  if (RVLSSpell.SpellChecker.SpellerType=sptISpell) and
    RVLSSpell.IsCyrillic(Language) then
    ConvertWinToKoi8(TRVRawByteString(Word2));
  if RVLSSpell.SpellChecker.IsKnownWord(Word2, Language) then
    exit;
  SelectCurrentWord;
  IsFormCreated := frm=nil;
  CreateSpellForm(TRVSpellFormStyle(RVLSSpell.SpellFormStyle), frm);
  if IsFormCreated and Assigned(RVLSSpell.OnSpellForm) then
    RVLSSpell.OnSpellForm(RVLSSpell, frm);
  SuggestionsStrings := frm.GetStringsForSuggestions;
  SuggestionsStrings.BeginUpdate;
  try
    SuggestionsStrings.Clear;
    RVLSSpell.SpellChecker.GetVariants(Word2, SuggestionsStrings, Language);
    if (RVLSSpell.SpellChecker.SpellerType=sptISpell) and
      RVLSSpell.IsCyrillic(Language) then
      for i := 0 to SuggestionsStrings.Count-1 do begin
        Word2 := SuggestionsStrings[i];
        ConvertKoi8ToWin(TRVRawByteString(Word2));
        SuggestionsStrings[i] := Word2;
      end;
  finally
    SuggestionsStrings.EndUpdate;
  end;
  BeforeShowSpellForm(FWordA, FTextA, FStartOffs, FEndOffs, frm,
    FRVData.GetRVStyle.TextStyles[FRVData.GetItemStyle(FItemNo)]);
  FDialogShown := True;
  mr := frm.ShowModal;
  Correction := frm.GetCorrectedText;
  RVLSSpell.DoSpellFormAction(mr, FWordA, Correction);
  case mr of
    mrNo: ;
    mrNoToAll:
      RVLSSpell.FIgnoreAll.Add(FWordA, Language);
    mrOk:
      RVLSSpell.SpellChecker.AddWord(FWordA, Language);
    mrYes:
      begin
        if not frm.IsReplacingItemText then begin
          FWordChanged := True;
          FWordA := Correction;
          end
        else begin
          FTextChanged := True;
          FTextA := Correction;
        end;
      end;
    mrYesToAll:
      begin
        RVLSSpell.FReplaceAll.Add(FWordA, Correction, Language);
        FWordChanged := True;
        FWordA := Correction;
      end;
    mrCancel:
      Result := False;
  end;
end;
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVLSSpell]);
end;


end.