
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVPolarSpellChecker: component for             }
{       spell-checking RichViewEdit using               }
{       Polar Spell Checker Component                   }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{       Polar SpellChecker Component:                   }
{        Copyright (c) Polar                            }
{       http://www.polarsoftware.com/                   }
{                                                       }
{*******************************************************}

{

  Installing this component:
  1) Download and install Polar Spell Checker Component (we used version 5.0)
  2) Menu Component | Import ActiveX Control, import and install
     "Polar SpellChecker 5.0 Component"
  3) Add this file (RVPolarSpell) in some package. Install this package.
      TRVPolarSpellChecker component will appear on "RichView" page of the
      Component Palette. This component should be used instead of TSpellChecker.

  Using:
    - place TRVPolarSpellChecker on the form.
    - in program, call TRVPolarSpellChecker.CheckRichViewEdit(rve, rvesFromCursor)
      to check rve: TRichViewEdit from the cursor position.
      Call TRVPolarSpellChecker.Check(rve, rvesFromStart) to check from the
      beginning.
  Functions for live spelling:
    - checking word in RichViewEdit.OnSpellingCheck:
      TRVPolarSpellChecker.IsWordIgnored and TRVPolarSpellChecker.DoesWordExist
    - adding to ignore list: TRVPolarSpellChecker.AddToIgnoreAll
    - adding to dictionary: TRVPolarSpellChecker.AddWord
    - suggestions: TRVPolarSpellChecker.GetSuggestionCount and
      TRVPolarSpellChecker.GetSuggestion
    - event on adding word to ignore list: TRVPolarSpellChecker.OnWordAddedToIgnore
    - event on adding word to dictionary: TRVPolarSpellChecker.OWordAdded
  Autocorrection:
    Call TRVPolarSpellChecker.AutoCorrectRichViewEdit when the user presses
      space, enter or tab character.

  Notes about the implementation:
  - this spell checker has a Unicode programming interface, so a complete
    multilanguage spelling support is possible only in Delphi 2009+
  - this spell checker does not have methods for accessing IgnoreAll list
    (necessary for live spelling); because of this, IgnoreAll list is duplicated
    in the component. This list is Unicode in Delphi 2005+.
}

unit RVPolarSpell;

interface

{$I RV_Defs.inc}

uses Windows, Classes,
     RVEdit, RVTypes, RVUni, RVFuncs,
     RVWordEnum, RVCustomSpellFrm,
     SPELLCHECKERLib_TLB;

type     
  {$IFDEF RICHVIEWDEF2005}
  {$IFNDEF RICHVIEWDEF2009}
    {$DEFINE USEWIDESTRINGLISTS}
  {$ENDIF}
  {$ENDIF}

  {$IFDEF USEWIDESTRINGLISTS}
  TRVPolSpellStringList = TWideStringList;
  {$ELSE}
  TRVPolSpellStringList = TStringList;
  {$ENDIF}

  TRVPolarSpellChecker = class (TSpellChecker)
  private
    // TSpellChecker does not provide access to ignore list, so implementing its copy
    FIgnoreList: TRVPolSpellStringList;
    FOnIgnoreAllCleared: TSpellCheckerIgnoreAllCleared;
    FOnWordAddedToIgnore: TSpellCheckerWordAddedToIgnore;
    procedure DoOnWordAddedToIgnore(ASender: TObject; const Object_: ISpellChecker;
      const Word: WideString; AlreadyExists: WordBool);
    procedure DoOnIgnoreAllCleared(ASender: TObject; const Object_: ISpellChecker);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CheckRichViewEdit(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
    function IsWordIgnored(const Word: TRVUnicodeString): Boolean;
    function AutoCorrectRichViewEdit(Editor: TCustomRichViewEdit): Boolean;
  published
    property OnWordAddedToIgnore: TSpellCheckerWordAddedToIgnore read FOnWordAddedToIgnore write FOnWordAddedToIgnore;
    property OnIgnoreAllCleared: TSpellCheckerIgnoreAllCleared read FOnIgnoreAllCleared write FOnIgnoreAllCleared;
  end;

  TRVPolarSpellCheckerHelper = class (TRVWordEnumerator)
    protected
      function ProcessWord: Boolean; override;
    public
      SpellChecker: TRVPolarSpellChecker;
  end;

procedure Register;

implementation

function IsAllCaps(const Word: TRVUnicodeString): Boolean;
var i: Integer;
begin
  Result := False;
  if not RVNT then
    exit;
  for i := 1 to Length(Word) do
    if not IsCharUpperW(Word[i]) then
      exit;
  Result := True;
end;
{------------------------------------------------------------------------------}
function HasNumbers(const Word: TRVUnicodeString): Boolean;
var i: Integer;
begin
  Result := True;
  for i := 1 to Length(Word) do
    if (ord(Word[i])>=ord('0')) and (ord(Word[i])<=ord('9')) then
      exit;
  Result := False;
end;
{=========================== TRVPolarSpellChecker =============================}
constructor TRVPolarSpellChecker.Create(AOwner: TComponent);
begin
  inherited;
  inherited OnIgnoreAllCleared := DoOnIgnoreAllCleared;
  inherited OnWordAddedToIgnore := DoOnWordAddedToIgnore;
  FIgnoreList := TRVPolSpellStringList.Create;
  {$IFDEF RICHVIEWDEF6}
  FIgnoreList.Sorted := True;
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
destructor TRVPolarSpellChecker.Destroy;
begin
  FIgnoreList.Free;
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TRVPolarSpellChecker.CheckRichViewEdit(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var Helper: TRVPolarSpellCheckerHelper;
begin
  Helper := TRVPolarSpellCheckerHelper.Create;
  try
    Helper.SpellChecker := Self;
    Helper.Unicode := True;
    Helper.Run(Editor, Scope);
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVPolarSpellChecker.DoOnIgnoreAllCleared(ASender: TObject;
  const Object_: ISpellChecker);
begin
  FIgnoreList.Clear;
  if Assigned(FOnIgnoreAllCleared) then
    FOnIgnoreAllCleared(ASender, Object_);
end;
{------------------------------------------------------------------------------}
procedure TRVPolarSpellChecker.DoOnWordAddedToIgnore(ASender: TObject;
  const Object_: ISpellChecker; const Word: WideString;
  AlreadyExists: WordBool);
begin
  FIgnoreList.Add(Word);
  if Assigned(FOnWordAddedToIgnore) then
    OnWordAddedToIgnore(ASender, Object_, Word, AlreadyExists);
end;
{------------------------------------------------------------------------------}
function TRVPolarSpellChecker.IsWordIgnored(
  const Word: TRVUnicodeString): Boolean;
{$IFDEF RICHVIEWDEF6}
var I: Integer;
{$ENDIF}
begin
  Result :=
    (IgnoreWordsInUppercase and IsAllCaps(Word)) or
    (IgnoreWordsWithNumbers and HasNumbers(Word));
  if Result then
    exit;
  {$IFDEF RVUNICODESTR}
  Result := RVIsEmail(Word) or RVIsURL(Word);
  {$ELSE}
  Result := RVIsEmail(String(Word)) or RVIsURL(String(Word));
  {$ENDIF}
  if Result then
    exit;
  {$IFDEF RICHVIEWDEF6}
  Result := FIgnoreList.Find(Word, I);
  {$ELSE}
  Result := FIgnoreList.IndexOf(Word)>=0;
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
function TRVPolarSpellChecker.AutoCorrectRichViewEdit(Editor: TCustomRichViewEdit): Boolean;
var ItemNo, WordEnd, WordStart: Integer;
    s,s2: TRVUnicodeString;
    OutText: OleVariant;
begin
  Result := False;
  if not AutoCorrectTextAsYouType then
    exit;
  Editor := Editor.TopLevelEditor;
  ItemNo := Editor.CurItemNo;
  WordEnd := Editor.OffsetInCurItem;
  if (Editor.GetItemStyle(ItemNo)<0) or
    (Editor.Style.TextStyles[Editor.GetItemStyle(ItemNo)].Charset=SYMBOL_CHARSET) then
    exit;
  s := Editor.GetItemTextW(ItemNo);
  WordStart := WordEnd;
  if WordStart<1 then
    exit;
  while (WordStart-1>0) and not Editor.RVData.IsDelimiterW(s[WordStart-1]) do
    dec(WordStart);
  s := Copy(s, WordStart, WordEnd-WordStart);
  CorrectText(s, OutText, False);
  s2 := OutText;
  Result := s<>s2;
  if Result then begin
    Editor.SetSelectionBounds(ItemNo, WordStart, ItemNo, WordEnd);
    Editor.InsertTextW(s2);
  end;
end;
{========================= TRVPolarSpellCheckerHelper =========================}
function TRVPolarSpellCheckerHelper.ProcessWord: Boolean;
var Res: TRVUnicodeString;
    Status: OleVariant;
begin
  Res := SpellChecker.CheckText(FWordW, Status);
  Result := Status=scrCheckingComplete;
  if not Result then
    exit;
  FWordChanged := Res<>FWordW;
  if FWordChanged then
    FWordW := Res;
end;
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVPolarSpellChecker]);
end;

end.
