
{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRVVSSpell3Helper: component for                }
{       spell-checking RichViewEdit using VSSpell 3.0   }
{       by Visual Components, Inc                       }
{                                                       }
{       RichView: Copyright (c) Sergey Tkachenko        }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{       VSSpell : Copyright (c) Visual Components, Inc  }
{                                                       }
{*******************************************************}

{
  VSSpell 3.0 is an ActiveX control shipped with Delphi.
  It is installed on "ActiveX" page of the Component Palette.

  Installing this component:
    - Add this file (RVVCVSSpell) in some package. Compile this package.
      TRVVSSpell3Helper component will appear on "RichView" page of the
      Component Palette.

  Using:
    - place TVSSpell and TRVVSSpell3Helper on the form.
    - in the Object Inspector, assign RVVSSpell3Helper.VSSpell to TVSSpell
    - in program, call RVVSSpell3Helper.Check(rve, rvesFromCursor) to check
      rve: TRichViewEdit from the cursor position.
      Call RVVSSpell3Helper.Check(rve, rvesFromStart) to check from the
      beginning.

  Known problems:
    - when user orders to replace some word with the empty string, replacement
      does not occur.

  Notes: I tried to implement this spell-checker via a component inherited from
    TVSSpell, but it crashes.

  Functions for live spelling:
    - checking word in RichViewEdit.OnSpellingCheck: TRVVSSpell3Helper.IsMisspelled
    - adding to ignore list: assign TRVVSSpell3Helper.VSSpell.MisspelledWord,
      then call TRVVSSpell3Helper.VSSpell.AddToCommonIRList or
      TRVVSSpell3Helper.VSSpell.AddToStandardIRList
    - adding to dictionary: assign TRVVSSpell3Helper.VSSpell.MisspelledWord,
      then call TRVVSSpell3Helper.VSSpell.AddToCustom
    - suggestions: TRVVSSpell3Helper.SuggestWords
    - event on adding word to ignore list: ?
    - event on adding word to dictionary: ?

  History:
  2011-03-20: property DialogShown specifies if a spelling check dialog has been
    shown during the last spelling check.
  2004-08-02: methods for live spelling check
}


unit RVVCVSSpell;

interface

uses SysUtils, Classes, Forms,
     RVEdit,
     RVWordEnum,
     VCSpell3;

type


  TRVVSSpell3HelperObject = class (TRVWordEnumerator)
    private
      FVSSpell: TVSSpell;
    protected
      function ProcessWord: Boolean; override;
    public
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
    published
      property VSSpell: TVSSpell read FVSSpell write FVSSpell;
  end;

  TRVVSSpell3Helper = class (TComponent)
    private
      FVSSpell: TVSSpell;
      FDialogShown: Boolean;
    protected
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    public
      procedure Check(Editor: TCustomRichViewEdit; Scope: TRVEnumScope);
      function IsMisspelled(const AWord: String): Boolean;
      procedure SuggestWords(const AWord: String; ASuggestions: TStrings);
      property DialogShown: Boolean read FDialogShown;      
    published
      property VSSpell: TVSSpell read FVSSpell write FVSSpell;
  end;



procedure Register;

implementation

{======================== TRVVSSpell3HelperObject =============================}
procedure TRVVSSpell3HelperObject.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
begin
  if FVSSpell=nil then
    raise Exception.Create('VSSpell is not assigned');
  Run(Editor, Scope);
end;
{------------------------------------------------------------------------------}
function TRVVSSpell3HelperObject.ProcessWord: Boolean;
var ResCode: Integer;
begin
  FVSSpell.AutoPopup := False;
  FVSSpell.AutoReplace := False;
  FVSSpell.CheckText := FWordW;
  ResCode := FVSSpell.ResultCode;
  while ResCode<0 do begin
    case ResCode of
      VSR_WORD_MISSPELLED, VSIR_IGNORE:
        begin
          SelectCurrentWord;
          FDialogShown := True;
          FVSSpell.PopupWordMisspelled  := 1;      // display misspell dialog box
          ResCode := FVSSpell.ResultCode;          // get new result
          if (ResCode<>VSR_CHECK_CANCELED) and
             (FVSSpell.ReplacementWord<>FWordW) and
             (FVSSpell.ReplacementWord<>'') then begin
            FWordW := FVSSpell.ReplacementWord;
            FWordChanged := True;
          end;
        end;
      VSR_BREAK:
        Application.ProcessMessages;
      VSR_IGNORE_REPLACE:
        if (FVSSpell.IRAction and VSIR_REPLACE_ALL)<>0 then begin
          FWordW := FVSSpell.ReplacementWord;
          FWordChanged := True;
        end;
    end;
    if ResCode=VSR_CHECK_CANCELED then begin
      Result := False;
      exit;
    end;
    ResCode := FVSSpell.ResumeCheck;
  end;
  Result := True;
end;
{========================= TRVVSSpell3Helper ==================================}
procedure TRVVSSpell3Helper.Check(Editor: TCustomRichViewEdit;
  Scope: TRVEnumScope);
var Helper: TRVVSSpell3HelperObject;
begin
  Helper := TRVVSSpell3HelperObject.Create;
  try
    Helper.VSSpell := VSSpell;
    Helper.Unicode := True;
    Helper.Check(Editor, Scope);
    FDialogShown := Helper.DialogShown;
  finally
    Helper.Free;
  end;
end;
{------------------------------------------------------------------------------}
function TRVVSSpell3Helper.IsMisspelled(const AWord: String): Boolean;
var ResCode: Integer;
begin
  Result := False;
  FVSSpell.AutoPopup := False;
  FVSSpell.AutoReplace := False;
  FVSSpell.CheckText := AWord;
  ResCode := FVSSpell.ResultCode;
  while ResCode<0 do begin
    if ResCode=VSR_WORD_MISSPELLED then
      Result := True;
    ResCode := FVSSpell.ResumeCheck;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell3Helper.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation=opRemove) and (AComponent=FVSSpell) then
    FVSSpell := nil;
end;
{------------------------------------------------------------------------------}
procedure TRVVSSpell3Helper.SuggestWords(const AWord: String;
  ASuggestions: TStrings);
var i: Integer;
begin
  ASuggestions.BeginUpdate;
  try
    ASuggestions.Clear;
    FVSSpell.FindSuggestions := AWord;
    for i := 1 to FVSSpell.SuggestionCount do
      ASuggestions.Add(FVSSpell.Suggestion[i]);
  finally
    ASuggestions.EndUpdate;
  end;
end;
{==============================================================================}
procedure Register;
begin
  RegisterComponents('RichView', [TRVVSSpell3Helper]);
end;

end.